"""
Graph module
"""

from collections import Counter

from .topics import Topics


# pylint: disable=R0904
class Graph:
    """
    Base class for Graph instances. This class builds graph networks. Supports topic modeling
    and relationship traversal.
    """

    def __init__(self, config):
        """
        Creates a new Graph.

        Args:
            config: graph configuration
        """

        # Graph configuration
        self.config = config if config is not None else {}

        # Graph backend
        self.backend = None

        # Topic modeling
        self.categories = None
        self.topics = None

        # Transform columns
        columns = config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        # Attributes to copy - skips text/object/relationship fields - set to True to copy all
        self.copyattributes = config.get("copyattributes", False)

        # Relationships are manually-provided edges
        self.relationships = columns.get("relationships", "relationships")
        self.relations = {}

    def create(self):
        """
        Creates the graph network.
        """

        raise NotImplementedError

    def count(self):
        """
        Returns the total number of nodes in graph.

        Returns:
            total nodes in graph
        """

        raise NotImplementedError

    def scan(self, attribute=None, data=False):
        """
        Iterates over nodes that match a criteria. If no criteria specified, all nodes
        are returned.

        Args:
            attribute: if specified, nodes having this attribute are returned
            data: if True, attribute data is also returned

        Returns:
            node id iterator if data is False or (id, attribute dictionary) iterator if data is True
        """

        raise NotImplementedError

    def node(self, node):
        """
        Get node by id. Returns None if not found.

        Args:
            node: node id

        Returns:
            graph node
        """

        raise NotImplementedError

    def addnode(self, node, **attrs):
        """
        Adds a node to the graph.

        Args:
            node: node id
            attrs: node attributes
        """

        raise NotImplementedError

    def addnodes(self, nodes):
        """
        Adds nodes to the graph.

        Args:
            nodes: list of (node, attributes) to add
        """

        raise NotImplementedError

    def removenode(self, node):
        """
        Removes a node and all it's edges from graph.

        Args:
            node: node id
        """

        raise NotImplementedError

    def hasnode(self, node):
        """
        Returns True if node found, False otherwise.

        Args:
            node: node id

        Returns:
            True if node found, False otherwise
        """

        raise NotImplementedError

    def attribute(self, node, field):
        """
        Gets a node attribute.

        Args:
            node: node id
            field: attribute name

        Returns:
            attribute value
        """

        raise NotImplementedError

    def addattribute(self, node, field, value):
        """
        Adds an attribute to node.

        Args:
            node: node id
            field: attribute name
            value: attribute value
        """

        raise NotImplementedError

    def removeattribute(self, node, field):
        """
        Removes an attribute from node.

        Args:
            node: node id
            field: attribute name

        Returns:
            attribute value or None if not present
        """

        raise NotImplementedError

    def edgecount(self):
        """
        Returns the total number of edges.

        Returns:
            total number of edges in graph
        """

        raise NotImplementedError

    def edges(self, node):
        """
        Gets edges of node by id.

        Args:
            node: node id

        Returns:
            list of edge node ids
        """

        raise NotImplementedError

    def addedge(self, source, target, **attrs):
        """
        Adds an edge to graph.

        Args:
            source: node 1 id
            target: node 2 id
        """

        raise NotImplementedError

    def addedges(self, edges):
        """
        Adds an edge to graph.

        Args:
            edges: list of (source, target, attributes) to add
        """

        raise NotImplementedError

    def hasedge(self, source, target=None):
        """
        Returns True if edge found, False otherwise. If target is None, this method
        returns True if any edge is found.

        Args:
            source: node 1 id
            target: node 2 id

        Returns:
            True if edge found, False otherwise
        """

        raise NotImplementedError

    def centrality(self):
        """
        Runs a centrality algorithm on the graph.

        Returns:
            dict of {node id: centrality score}
        """

        raise NotImplementedError

    def pagerank(self):
        """
        Runs the pagerank algorithm on the graph.

        Returns:
            dict of {node id, page rank score}
        """

        raise NotImplementedError

    def showpath(self, source, target):
        """
        Gets the shortest path between source and target.

        Args:
            source: start node id
            target: end node id

        Returns:
            list of node ids representing the shortest path
        """

        raise NotImplementedError

    def isquery(self, queries):
        """
        Checks if queries are supported graph queries.

        Args:
            queries: queries to check

        Returns:
            True if all the queries are supported graph queries, False otherwise
        """

        raise NotImplementedError

    def parse(self, query):
        """
        Parses a graph query into query components.

        Args:
            query: graph query

        Returns:
            query components as a dictionary
        """

        raise NotImplementedError

    def search(self, query, limit=None, graph=False):
        """
        Searches graph for nodes matching query.

        Args:
            query: graph query
            limit: maximum results
            graph: return graph results if True

        Returns:
            list of dict if graph is set to False
            filtered graph if graph is set to True
        """

        raise NotImplementedError

    def batchsearch(self, queries, limit=None, graph=False):
        """
        Searches graph for nodes matching query.

        Args:
            query: graph query
            limit: maximum results
            graph: return graph results if True

        Returns:
            list of dict if graph is set to False
            filtered graph if graph is set to True
        """

        return [self.search(query, limit, graph) for query in queries]

    def communities(self, config):
        """
        Run community detection on the graph.

        Args:
            config: configuration

        Returns:
            dictionary of {topic name:[ids]}
        """

        raise NotImplementedError

    def load(self, path):
        """
        Loads a graph at path.

        Args:
            path: path to graph
        """

        raise NotImplementedError

    def save(self, path):
        """
        Saves a graph at path.

        Args:
            path: path to save graph
        """

        raise NotImplementedError

    def loaddict(self, data):
        """
        Loads data from input dictionary into this graph.

        Args:
            data: input dictionary
        """

        raise NotImplementedError

    def savedict(self):
        """
        Saves graph data to a dictionary.

        Returns:
            dict
        """

        raise NotImplementedError

    def initialize(self):
        """
        Initialize graph instance.
        """

        if not self.backend:
            self.backend = self.create()

    def close(self):
        """
        Closes this graph.
        """

        self.backend, self.categories, self.topics = None, None, None

    def insert(self, documents, index=0):
        """
        Insert graph nodes for each document.

        Args:
            documents: list of (id, data, tags)
            index: indexid offset, used for node ids
        """

        # Initialize graph backend
        self.initialize()

        nodes = []
        for uid, document, _ in documents:
            # Manually provided relationships and attributes to copy
            relations, attributes = None, {}

            # Extract data from dictionary
            if isinstance(document, dict):
                # Extract relationships
                relations = document.get(self.relationships)

                # Attributes to copy, if any
                search = self.copyattributes if isinstance(self.copyattributes, list) else []
                attributes = {
                    k: v
                    for k, v in document.items()
                    if k not in [self.text, self.object, self.relationships] and (self.copyattributes is True or k in search)
                }

                # Require text or object field
                document = document.get(self.text, document.get(self.object))

            if document is not None:
                if isinstance(document, list):
                    # Join tokens as text
                    document = " ".join(document)

                # Create node
                nodes.append((index, {**{"id": uid, "data": document}, **attributes}))

                # Add relationships
                self.addrelations(index, relations)

                index += 1

        # Add nodes
        self.addnodes(nodes)

    def delete(self, ids):
        """
        Deletes ids from graph.

        Args:
            ids: node ids to delete
        """

        for node in ids:
            # Remove existing node, if it exists
            if self.hasnode(node):
                # Delete from topics
                topic = self.attribute(node, "topic")
                if topic and self.topics:
                    # Delete id from topic
                    self.topics[topic].remove(node)

                    # Also delete topic, if it's empty
                    if not self.topics[topic]:
                        self.topics.pop(topic)

                # Delete node
                self.removenode(node)

    def index(self, search, ids, similarity):
        """
        Build relationships between graph nodes using a score-based search function.

        Args:
            search: batch search function - takes a list of queries and returns lists of (id, scores) to use as edge weights
            ids: ids function - internal id resolver
            similarity: batch similarity function - takes a list of text and labels and returns best matches
        """

        # Add relationship edges
        self.resolverelations(ids)

        # Infer node edges using search function
        self.inferedges(self.scan(), search)

        # Label categories/topics
        if "topics" in self.config:
            self.addtopics(similarity)

    def upsert(self, search, ids, similarity=None):
        """
        Adds relationships for new graph nodes using a score-based search function.

        Args:
            search: batch search function - takes a list of queries and returns lists of (id, scores) to use as edge weights
            ids: ids function - internal id resolver
            similarity: batch similarity function - takes a list of text and labels and returns best matches
        """

        # Detect if topics processing is enabled
        hastopics = "topics" in self.config

        # Add relationship edges
        self.resolverelations(ids)

        # Infer node edges using new/updated nodes, set updated flag for topic processing, if necessary
        self.inferedges(self.scan(attribute="data"), search, {"updated": True} if hastopics else None)

        # Infer topics with topics of connected nodes
        if hastopics:
            # Infer topics if there is at least one topic, otherwise rebuild
            if self.topics:
                self.infertopics()
            else:
                self.addtopics(similarity)

    def filter(self, nodes, graph=None):
        """
        Creates a subgraph of this graph using the list of input nodes. This method creates a new graph
        selecting only matching nodes, edges, topics and categories.

        Args:
            nodes: nodes to select as a list of ids or list of (id, score) tuples
            graph: optional graph used to store filtered results

        Returns:
            graph
        """

        # Set graph if available, otherwise create a new empty graph of the same type
        graph = graph if graph else type(self)(self.config)

        # Initalize subgraph
        graph.initialize()

        nodeids = {node[0] if isinstance(node, tuple) else node for node in nodes}
        for node in nodes:
            # Unpack node and score, if available
            node, score = node if isinstance(node, tuple) else (node, None)

            # Add nodes
            graph.addnode(node, **self.node(node))

            # Add score if present
            if score is not None:
                graph.addattribute(node, "score", score)

            # Add edges
            edges = self.edges(node)
            if edges:
                for target, attributes in self.edges(node).items():
                    if target in nodeids:
                        graph.addedge(node, target, **attributes)

        # Filter categories and topics
        if self.topics:
            topics = {}
            for i, (topic, ids) in enumerate(self.topics.items()):
                ids = [x for x in ids if x in nodeids]
                if ids:
                    topics[topic] = (self.categories[i] if self.categories else None, ids)

            # Sort by number of nodes descending
            topics = sorted(topics.items(), key=lambda x: len(x[1][1]), reverse=True)

            # Copy filtered categories and topics
            graph.categories = [category for _, (category, _) in topics] if self.categories else None
            graph.topics = {topic: ids for topic, (_, ids) in topics}

        return graph

    def addrelations(self, node, relations):
        """
        Add manually-provided relationships.

        Args:
            node: node id
            relations: list of relationships to add
        """

        # Add relationships, if any
        if relations:
            if node not in self.relations:
                self.relations[node] = []

            # Add each relationship
            for relation in relations:
                # Support both dict and string ids
                relation = {"id": relation} if not isinstance(relation, dict) else relation
                self.relations[node].append(relation)

    def resolverelations(self, ids):
        """
        Resolves ids and creates edges for manually-provided relationships.

        Args:
            ids: internal id resolver
        """

        # Relationship edges
        edges = []

        # Resolve ids and create edges for relationships
        for node, relations in self.relations.items():
            # Resolve internal ids
            iids = ids(y["id"] for y in relations)

            # Add each edge
            for relation in relations:
                # Make copy of relation
                relation = relation.copy()

                # Lookup targets for relationship
                targets = iids.get(str(relation.pop("id")))

                # Create edge for each instance of id - internal id pair
                if targets:
                    for target in targets:
                        # Add weight, if not provided
                        relation["weight"] = relation.get("weight", 1.0)

                        # Add edge and all other attributes
                        edges.append((node, target, relation))

        # Add relationships
        if edges:
            self.addedges(edges)

        # Clear temporary relationship storage
        self.relations = {}

    def inferedges(self, nodes, search, attributes=None):
        """
        Infers edges for a list of nodes using a score-based search function.

        Args:
            nodes: list of nodes
            search: search function to use to identify edges
            attribute: dictionary of attributes to add to each node
        """

        # Read graph parameters
        batchsize, limit, minscore = self.config.get("batchsize", 256), self.config.get("limit", 15), self.config.get("minscore", 0.1)
        approximate = self.config.get("approximate", True)

        batch = []
        for node in nodes:
            # Get data attribute
            data = self.removeattribute(node, "data")

            # Set text field when data is a string
            if isinstance(data, str):
                self.addattribute(node, "text", data)

            # Add additional attributes, if specified
            if attributes:
                for field, value in attributes.items():
                    self.addattribute(node, field, value)

            # Skip nodes with existing edges when building an approximate network
            if not approximate or not self.hasedge(node):
                batch.append((node, data))

            # Process batch
            if len(batch) == batchsize:
                self.addbatch(search, batch, limit, minscore)
                batch = []

        if batch:
            self.addbatch(search, batch, limit, minscore)

    def addbatch(self, search, batch, limit, minscore):
        """
        Adds batch of documents to graph. This method runs the search function for each item in batch
        and adds node edges between the input and each search result.

        Args:
            search: search function to use to identify edges
            batch: batch to add
            limit: max edges to add per node
            minscore: min score to add node edge
        """

        edges = []
        for x, result in enumerate(search([data for _, data in batch], limit)):
            # Get input node id
            x, _ = batch[x]

            # Add edges for each input node id and result node id pair that meets specified criteria
            for y, score in result:
                if str(x) != str(y) and score > minscore:
                    edges.append((x, y, {"weight": score}))

        self.addedges(edges)

    def addtopics(self, similarity=None):
        """
        Identifies and adds topics using community detection.

        Args:
            similarity: similarity function for labeling categories
        """

        # Clear previous topics, if any
        self.cleartopics()

        # Use community detection to get topics
        topics = Topics(self.config["topics"])
        config = topics.config
        self.topics = topics(self)

        # Label each topic with a higher level category
        if "categories" in config and similarity:
            self.categories = []
            results = similarity(self.topics.keys(), config["categories"])
            for result in results:
                self.categories.append(config["categories"][result[0][0]])

        # Add topic-related node attributes
        for x, topic in enumerate(self.topics):
            for r, node in enumerate(self.topics[topic]):
                self.addattribute(node, "topic", topic)
                self.addattribute(node, "topicrank", r)

                if self.categories:
                    self.addattribute(node, "category", self.categories[x])

    def cleartopics(self):
        """
        Clears topic fields from all nodes.
        """

        # Clear previous topics, if any
        if self.topics:
            for node in self.scan():
                self.removeattribute(node, "topic")
                self.removeattribute(node, "topicrank")

                if self.categories:
                    self.removeattribute(node, "category")

            self.topics, self.categories = None, None

    def infertopics(self):
        """
        Infers topics for all nodes with an "updated" attribute. This method analyzes the direct node
        neighbors and set the most commonly occuring topic and category for each node.
        """

        # Iterate over nodes missing topic attribute (only occurs for new nodes)
        for node in self.scan(attribute="updated"):
            # Remove updated attribute
            self.removeattribute(node, "updated")

            # Get list of neighboring nodes
            ids = self.edges(node)
            ids = ids.keys() if ids else None

            # Infer topic
            topic = Counter(self.attribute(x, "topic") for x in ids).most_common(1)[0][0] if ids else None
            if topic:
                # Add id to topic list and set topic attribute
                self.topics[topic].append(node)
                self.addattribute(node, "topic", topic)

                # Set topic rank
                self.addattribute(node, "topicrank", len(self.topics[topic]) - 1)

                # Infer category
                category = Counter(self.attribute(x, "category") for x in ids).most_common(1)[0][0]
                self.addattribute(node, "category", category)



"""
Pipeline factory module
"""

import inspect
import sys
import types

from ..util import Resolver

from .base import Pipeline


class PipelineFactory:
    """
    Pipeline factory. Creates new Pipeline instances.
    """

    @staticmethod
    def get(pipeline):
        """
        Gets a new instance of pipeline class.

        Args:
            pclass: Pipeline instance class

        Returns:
            Pipeline class
        """

        # Local pipeline if no package
        if "." not in pipeline:
            return PipelineFactory.list()[pipeline]

        # Attempt to load custom pipeline
        return Resolver()(pipeline)

    @staticmethod
    def create(config, pipeline):
        """
        Creates a new Pipeline instance.

        Args:
            config: Pipeline configuration
            pipeline: Pipeline instance class

        Returns:
            Pipeline
        """

        # Resolve pipeline
        pipeline = PipelineFactory.get(pipeline)

        # Return functions directly, otherwise create pipeline instance
        return pipeline if isinstance(pipeline, types.FunctionType) else pipeline(**config)

    @staticmethod
    def list():
        """
        Lists callable pipelines.

        Returns:
            {short name: pipeline class}
        """

        pipelines = {}

        # Get handle to pipeline module
        pipeline = sys.modules[".".join(__name__.split(".")[:-1])]

        # Get list of callable pipelines
        for x in inspect.getmembers(pipeline, inspect.isclass):
            if issubclass(x[1], Pipeline) and [y for y, _ in inspect.getmembers(x[1], inspect.isfunction) if y == "__call__"]:
                # short name: pipeline class
                pipelines[x[0].lower()] = x[1]

        return pipelines



"""
Database imports
"""

from .base import Database
from .client import Client
from .duckdb import DuckDB
from .embedded import Embedded
from .encoder import *
from .factory import DatabaseFactory
from .rdbms import RDBMS
from .schema import *
from .sqlite import SQLite
from .sql import *



"""
Task imports
"""

from .base import Task
from .console import ConsoleTask
from .export import ExportTask
from .factory import TaskFactory
from .file import FileTask
from .image import ImageTask
from .retrieve import RetrieveTask
from .service import ServiceTask
from .storage import StorageTask
from .stream import StreamTask
from .template import *
from .template import RagTask as ExtractorTask
from .url import UrlTask
from .workflow import WorkflowTask



"""
Factory module
"""

from ..util import Resolver

from .networkx import NetworkX
from .rdbms import RDBMS


class GraphFactory:
    """
    Methods to create graphs.
    """

    @staticmethod
    def create(config):
        """
        Create a Graph.

        Args:
            config: graph configuration

        Returns:
            Graph
        """

        # Graph instance
        graph = None
        backend = config.get("backend", "networkx")

        # Create graph instance
        if backend == "networkx":
            graph = NetworkX(config)
        elif backend == "rdbms":
            graph = RDBMS(config)
        else:
            graph = GraphFactory.resolve(backend, config)

        # Store config back
        config["backend"] = backend

        return graph

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            Graph
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve graph backend: '{backend}'") from e



"""
MessagePack module
"""

from io import BytesIO
from typing import Any

import msgpack

from fastapi import Response
from PIL.Image import Image


class MessagePackResponse(Response):
    """
    Encodes responses with MessagePack.
    """

    media_type = "application/msgpack"

    def render(self, content: Any) -> bytes:
        """
        Renders content to the response.

        Args:
            content: input content

        Returns:
            rendered content as bytes
        """

        return msgpack.packb(content, default=MessagePackEncoder())


class MessagePackEncoder:
    """
    Extended MessagePack encoder that converts images to bytes.
    """

    def __call__(self, o):
        # Convert Image to bytes
        if isinstance(o, Image):
            buffered = BytesIO()
            o.save(buffered, format=o.format, quality="keep")
            o = buffered

        # Get bytes from BytesIO
        if isinstance(o, BytesIO):
            o = o.getvalue()

        return o



"""
ORM Module
"""

# Conditional import
try:
    from sqlalchemy import Column, DateTime, Float, JSON, Integer, LargeBinary, String, Text
    from sqlalchemy.orm import DeclarativeBase

    ORM = True
except ImportError:
    ORM = False


# Standard database schema using object relational mapping (ORM).
if ORM:

    def idcolumn():
        """
        Creates an id column. This method creates an unbounded text field for platforms that support it.

        Returns:
            id column definition
        """

        return String(512).with_variant(Text(), "sqlite", "postgresql")

    class Base(DeclarativeBase):
        """
        Base mapping.
        """

    class Batch(Base):
        """
        Batch temporary table mapping.
        """

        __tablename__ = "batch"
        __table_args__ = {"prefixes": ["TEMPORARY"]}

        autoid = Column(Integer, primary_key=True, autoincrement=True)
        indexid = Column(Integer)
        id = Column(idcolumn())
        batch = Column(Integer)

    class Score(Base):
        """
        Scores temporary table mapping.
        """

        __tablename__ = "scores"
        __table_args__ = {"prefixes": ["TEMPORARY"]}

        indexid = Column(Integer, primary_key=True, autoincrement=False)
        score = Column(Float)

    class Document(Base):
        """
        Documents table mapping.
        """

        __tablename__ = "documents"

        id = Column(idcolumn(), primary_key=True)
        data = Column(JSON)
        tags = Column(Text)
        entry = Column(DateTime(timezone=True))

    class Object(Base):
        """
        Objects table mapping.
        """

        __tablename__ = "objects"

        id = Column(idcolumn(), primary_key=True)
        object = Column(LargeBinary)
        tags = Column(Text)
        entry = Column(DateTime(timezone=True))

    class SectionBase(Base):
        """
        Generic sections table mapping. Allows multiple section table names for reindexing.
        """

        __abstract__ = True

        indexid = Column(Integer, primary_key=True, autoincrement=False)
        id = Column(idcolumn(), index=True)
        text = Column(Text)
        tags = Column(Text)
        entry = Column(DateTime(timezone=True))

    class Section(SectionBase):
        """
        Section table mapping.
        """

        __tablename__ = "sections"



"""
Client module
"""

import os
import time

# Conditional import
try:
    from sqlalchemy import StaticPool, Text, cast, create_engine, insert, text as textsql
    from sqlalchemy.orm import Session, aliased
    from sqlalchemy.schema import CreateSchema

    from .schema import Base, Batch, Document, Object, Section, SectionBase, Score

    ORM = True
except ImportError:
    ORM = False

from .rdbms import RDBMS


class Client(RDBMS):
    """
    Database client instance. This class connects to an external database using SQLAlchemy. It supports any database
    that is supported by SQLAlchemy (PostgreSQL, MariaDB, etc) and has JSON support.
    """

    def __init__(self, config):
        """
        Creates a new Database.

        Args:
            config: database configuration parameters
        """

        super().__init__(config)

        if not ORM:
            raise ImportError('SQLAlchemy is not available - install "database" extra to enable')

        # SQLAlchemy parameters
        self.engine, self.dbconnection = None, None

    def save(self, path):
        # Commit session and database connection
        super().save(path)

        if self.dbconnection:
            self.dbconnection.commit()

    def close(self):
        super().close()

        # Dispose of engine, which also closes dbconnection
        if self.engine:
            self.engine.dispose()

    def reindexstart(self):
        # Working table name
        name = f"rebuild{round(time.time() * 1000)}"

        # Create working table metadata
        type("Rebuild", (SectionBase,), {"__tablename__": name})
        Base.metadata.tables[name].create(self.dbconnection)

        return name

    def reindexend(self, name):
        # Remove table object from metadata
        Base.metadata.remove(Base.metadata.tables[name])

    def jsonprefix(self):
        # JSON column prefix
        return "cast("

    def jsoncolumn(self, name):
        # Alias documents table
        d = aliased(Document, name="d")

        # Build JSON column expression for column
        return str(cast(d.data[name].as_string(), Text).compile(dialect=self.engine.dialect, compile_kwargs={"literal_binds": True}))

    def createtables(self):
        # Create tables
        Base.metadata.create_all(self.dbconnection, checkfirst=True)

        # Clear existing data - table schema is created upon connecting to database
        for table in ["sections", "documents", "objects"]:
            self.cursor.execute(f"DELETE FROM {table}")

    def finalize(self):
        # Flush cached objects
        self.connection.flush()

    def insertdocument(self, uid, data, tags, entry):
        self.connection.add(Document(id=uid, data=data, tags=tags, entry=entry))

    def insertobject(self, uid, data, tags, entry):
        self.connection.add(Object(id=uid, object=data, tags=tags, entry=entry))

    def insertsection(self, index, uid, text, tags, entry):
        # Save text section
        self.connection.add(Section(indexid=index, id=uid, text=text, tags=tags, entry=entry))

    def createbatch(self):
        # Create temporary batch table, if necessary
        Base.metadata.tables["batch"].create(self.dbconnection, checkfirst=True)

    def insertbatch(self, indexids, ids, batch):
        if indexids:
            self.connection.execute(insert(Batch), [{"indexid": i, "batch": batch} for i in indexids])
        if ids:
            self.connection.execute(insert(Batch), [{"id": str(uid), "batch": batch} for uid in ids])

    def createscores(self):
        # Create temporary scores table, if necessary
        Base.metadata.tables["scores"].create(self.dbconnection, checkfirst=True)

    def insertscores(self, scores):
        # Average scores by id
        if scores:
            self.connection.execute(insert(Score), [{"indexid": i, "score": sum(s) / len(s)} for i, s in scores.items()])

    def connect(self, path=None):
        # Connection URL
        content = self.config.get("content")

        # Read ENV variable, if necessary
        content = os.environ.get("CLIENT_URL") if content == "client" else content

        # Create engine using database URL
        self.engine = create_engine(content, poolclass=StaticPool, echo=False, json_serializer=lambda x: x)
        self.dbconnection = self.engine.connect()

        # Create database session
        database = Session(self.dbconnection)

        # Set default schema, if necessary
        schema = self.config.get("schema")
        if schema:
            with self.engine.begin():
                self.sqldialect(database, CreateSchema(schema, if_not_exists=True))

            self.sqldialect(database, textsql("SET search_path TO :schema"), {"schema": schema})

        return database

    def getcursor(self):
        return Cursor(self.connection)

    def rows(self):
        return self.cursor

    def addfunctions(self):
        return

    def sqldialect(self, database, sql, parameters=None):
        """
        Executes a SQL statement based on the current SQL dialect.

        Args:
            database: current database
            sql: SQL to execute
            parameters: optional bind parameters
        """

        args = (sql, parameters) if self.engine.dialect.name == "postgresql" else (textsql("SELECT 1"),)
        database.execute(*args)


class Cursor:
    """
    Implements basic compatibility with the Python DB-API.
    """

    def __init__(self, connection):
        self.connection = connection
        self.result = None

    def __iter__(self):
        return self.result

    def execute(self, statement, parameters=None):
        """
        Executes statement.

        Args:
            statement: statement to execute
            parameters: optional dictionary with bind parameters
        """

        if isinstance(statement, str):
            statement = textsql(statement)

        self.result = self.connection.execute(statement, parameters)

    def fetchall(self):
        """
        Fetches all rows from the current result.

        Returns:
            all rows from current result
        """

        return self.result.all() if self.result else None

    def fetchone(self):
        """
        Fetches first row from current result.

        Returns:
            first row from current result
        """

        return self.result.first() if self.result else None

    @property
    def description(self):
        """
        Returns columns for current result.

        Returns:
            list of columns
        """

        return [(key,) for key in self.result.keys()] if self.result else None



"""
Hugging Face module
"""

from threading import Thread

from transformers import AutoModelForImageTextToText, TextIteratorStreamer

from ...models import Models

from ..hfpipeline import HFPipeline

from .generation import Generation


class HFGeneration(Generation):
    """
    Hugging Face Transformers generative model.
    """

    def __init__(self, path, template=None, **kwargs):
        # Call parent constructor
        super().__init__(path, template, **kwargs)

        # Create HuggingFace LLM pipeline
        self.llm = HFLLM(path, **kwargs)

    def isvision(self):
        return isinstance(self.llm.pipeline.model, AutoModelForImageTextToText)

    def stream(self, texts, maxlength, stream, stop, **kwargs):
        yield from self.llm(texts, maxlength=maxlength, stream=stream, stop=stop, **kwargs)


class HFLLM(HFPipeline):
    """
    Hugging Face Transformers large language model (LLM) pipeline. This pipeline autodetects if the model path
    is a text generation or sequence to sequence model.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, task=None, **kwargs):
        super().__init__(self.task(path, task, **kwargs), path, quantize, gpu, model, **kwargs)

        # Load tokenizer, if necessary
        self.pipeline.tokenizer = self.pipeline.tokenizer if self.pipeline.tokenizer else Models.tokenizer(path, **kwargs)

    def __call__(self, text, prefix=None, maxlength=512, workers=0, stream=False, stop=None, **kwargs):
        """
        Generates text. Supports the following input formats:

          - String or list of strings (instruction-tuned models must follow chat templates)
          - List of dictionaries with `role` and `content` key-values or lists of lists

        Args:
            text: text|list
            prefix: optional prefix to prepend to text elements
            maxlength: maximum sequence length
            workers: number of concurrent workers to use for processing data, defaults to None
            stream: stream response if True, defaults to False
            stop: list of stop strings
            kwargs: additional generation keyword arguments

        Returns:
            generated text
        """

        # List of texts
        texts = text if isinstance(text, list) else [text]

        # Add prefix, if necessary
        if prefix:
            texts = [f"{prefix}{x}" for x in texts]

        # Combine all keyword arguments
        args, kwargs = self.parameters(texts, maxlength, workers, stop, **kwargs)

        # Stream response
        if stream:
            return StreamingResponse(self.pipeline, texts, stop, **kwargs)()

        # Run pipeline and extract generated text
        results = [self.extract(result) for result in self.pipeline(*args, **kwargs)]

        return results[0] if isinstance(text, str) else results

    def parameters(self, texts, maxlength, workers, stop, **kwargs):
        """
        Builds a list of arguments and a combined parameter dictionary to use as keyword arguments.

        Args:
            texts: input texts
            maxlength: maximum sequence length
            workers: number of concurrent workers to use for processing data, defaults to None
            stop: list of stop strings
            kwargs: additional generation keyword arguments

        Returns:
            args, kwargs
        """

        # Set defaults and get underlying model
        defaults, model = {"max_length": maxlength, "max_new_tokens": None, "num_workers": workers}, self.pipeline.model

        # Set parameters for vision models and return
        if self.pipeline.task == "image-text-to-text":
            # Maxlength has to be large enough to accomodate images
            defaults["max_length"] = max(maxlength, 2048)

            # Set default token id
            tokenid = model.generation_config.pad_token_id
            model.generation_config.pad_token_id = tokenid if tokenid else model.generation_config.eos_token_id

            # Vision models take all arguments as keyword arguments
            return [], {**{"text": texts, "truncation": True}, **defaults, **kwargs}

        # Add pad token if it's missing from model config
        if not model.config.pad_token_id:
            tokenid = model.config.eos_token_id
            tokenid = tokenid[0] if isinstance(tokenid, list) else tokenid

            # Set pad_token_id parameter
            defaults["pad_token_id"] = tokenid

            # Update tokenizer for batching
            if "batch_size" in kwargs and self.pipeline.tokenizer.pad_token_id is None:
                self.pipeline.tokenizer.pad_token_id = tokenid
                self.pipeline.tokenizer.padding_side = "left"

        # Set tokenizer when stop strings is set
        if stop:
            defaults["tokenizer"] = self.pipeline.tokenizer

        return [texts], {**defaults, **kwargs}

    def extract(self, result):
        """
        Extracts generated text from a pipeline result.

        Args:
            result: pipeline result

        Returns:
            generated text
        """

        # Extract output from list, if necessary
        result = result[0] if isinstance(result, list) else result
        text = result["generated_text"]
        return text[-1]["content"] if isinstance(text, list) else text

    def task(self, path, task, **kwargs):
        """
        Get the pipeline task name.

        Args:
            path: model path input
            task: task name
            kwargs: optional additional keyword arguments

        Returns:
            pipeline task name
        """

        # Mapping from txtai to Hugging Face pipeline tasks
        mapping = {"language-generation": "text-generation", "sequence-sequence": "text2text-generation", "vision": "image-text-to-text"}

        # Attempt to resolve task
        if path and not task:
            task = Models.task(path, **kwargs)

        # Map to Hugging Face task. Default to text2text-generation pipeline when task not resolved.
        return mapping.get(task, "text2text-generation")


class Generator(HFLLM):
    """
    Generate text with a causal language model.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        super().__init__(path, quantize, gpu, model, "language-generation", **kwargs)


class Sequences(HFLLM):
    """
    Generate text with a sequence-sequence model.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        super().__init__(path, quantize, gpu, model, "sequence-sequence", **kwargs)


class StreamingResponse:
    """
    Generate text as a streaming response.
    """

    def __init__(self, pipeline, texts, stop, **kwargs):
        # Create streamer
        self.stream = TextIteratorStreamer(pipeline.tokenizer, skip_prompt=True, skip_special_tokens=True, timeout=5)
        kwargs["streamer"] = self.stream
        kwargs["stop_strings"] = stop

        # Create thread
        self.thread = Thread(target=pipeline, args=[texts], kwargs=kwargs)

        # Store number of inputs
        self.length = len(texts)

    def __call__(self):
        # Start the process
        self.thread.start()

        return self

    def __iter__(self):
        for _ in range(self.length):
            yield from self.stream



"""
Defines API paths for entity endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/entity")
def entity(text: str):
    """
    Applies a token classifier to text.

    Args:
        text: input text

    Returns:
        list of (entity, entity type, score) per text element
    """

    return application.get().pipeline("entity", (text,))


@router.post("/batchentity")
def batchentity(texts: List[str] = Body(...)):
    """
    Applies a token classifier to text.

    Args:
        texts: list of text

    Returns:
        list of (entity, entity type, score) per text element
    """

    return application.get().pipeline("entity", (texts,))



"""
Similarity module
"""

import numpy as np

from .crossencoder import CrossEncoder
from .labels import Labels
from .lateencoder import LateEncoder


class Similarity(Labels):
    """
    Computes similarity between query and list of text using a transformers model.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, dynamic=True, crossencode=False, lateencode=False, **kwargs):
        self.crossencoder, self.lateencoder = None, None

        if lateencode:
            # Load a late interaction encoder if lateencode set to True
            self.lateencoder = LateEncoder(path=path, gpu=gpu, **kwargs)
        else:
            # Use zero-shot classification if dynamic is True and crossencode is False, otherwise use standard text classification
            super().__init__(path, quantize, gpu, model, False if crossencode else dynamic, **kwargs)

            # Load as a cross-encoder if crossencode set to True
            self.crossencoder = CrossEncoder(model=self.pipeline) if crossencode else None

    # pylint: disable=W0222
    def __call__(self, query, texts, multilabel=True, **kwargs):
        """
        Computes the similarity between query and list of text. Returns a list of
        (id, score) sorted by highest score, where id is the index in texts.

        This method supports query as a string or a list. If the input is a string,
        the return type is a 1D list of (id, score). If text is a list, a 2D list
        of (id, score) is returned with a row per string.

        Args:
            query: query text|list
            texts: list of text
            multilabel: labels are independent if True, scores are normalized to sum to 1 per text item if False, raw scores returned if None
            kwargs: additional keyword args

        Returns:
            list of (id, score)
        """

        if self.crossencoder:
            # pylint: disable=E1102
            return self.crossencoder(query, texts, multilabel)

        if self.lateencoder:
            return self.lateencoder(query, texts)

        # Call Labels pipeline for texts using input query as the candidate label
        scores = super().__call__(texts, [query] if isinstance(query, str) else query, multilabel, **kwargs)

        # Sort on query index id
        scores = [[score for _, score in sorted(row)] for row in scores]

        # Transpose axes to get a list of text scores for each query
        scores = np.array(scores).T.tolist()

        # Build list of (id, score) per query sorted by highest score
        scores = [sorted(enumerate(row), key=lambda x: x[1], reverse=True) for row in scores]

        return scores[0] if isinstance(query, str) else scores

    def encode(self, data, category):
        """
        Encodes a batch of data using the underlying model.

        Args:
            data: input data
            category: encoding category

        Returns:
            encoded data
        """

        return self.lateencoder.encode(data, category) if self.lateencoder else data



"""
Router imports
"""

from . import agent
from . import caption
from . import embeddings
from . import entity
from . import extractor
from . import labels
from . import llm
from . import objects
from . import openai
from . import rag
from . import reranker
from . import segmentation
from . import similarity
from . import summary
from . import tabular
from . import textractor
from . import texttospeech
from . import transcription
from . import translation
from . import workflow
from . import upload



"""
API module
"""

import json

from .cluster import Cluster

from ..app import Application


class API(Application):
    """
    Base API template. The API is an extended txtai application, adding the ability to cluster API instances together.

    Downstream applications can extend this base template to add/modify functionality.
    """

    def __init__(self, config, loaddata=True):
        super().__init__(config, loaddata)

        # Embeddings cluster
        self.cluster = None
        if self.config.get("cluster"):
            self.cluster = Cluster(self.config["cluster"])

    # pylint: disable=W0221
    def search(self, query, limit=None, weights=None, index=None, parameters=None, graph=False, request=None):
        # When search is invoked via the API, limit is set from the request
        # When search is invoked directly, limit is set using the method parameter
        limit = self.limit(request.query_params.get("limit") if request and hasattr(request, "query_params") else limit)
        weights = self.weights(request.query_params.get("weights") if request and hasattr(request, "query_params") else weights)
        index = request.query_params.get("index") if request and hasattr(request, "query_params") else index
        parameters = request.query_params.get("parameters") if request and hasattr(request, "query_params") else parameters
        graph = request.query_params.get("graph") if request and hasattr(request, "query_params") else graph

        # Decode parameters
        parameters = json.loads(parameters) if parameters and isinstance(parameters, str) else parameters

        if self.cluster:
            return self.cluster.search(query, limit, weights, index, parameters, graph)

        return super().search(query, limit, weights, index, parameters, graph)

    def batchsearch(self, queries, limit=None, weights=None, index=None, parameters=None, graph=False):
        if self.cluster:
            return self.cluster.batchsearch(queries, self.limit(limit), weights, index, parameters, graph)

        return super().batchsearch(queries, limit, weights, index, parameters, graph)

    def add(self, documents):
        """
        Adds a batch of documents for indexing.

        Downstream applications can override this method to also store full documents in an external system.

        Args:
            documents: list of {id: value, text: value}

        Returns:
            unmodified input documents
        """

        if self.cluster:
            self.cluster.add(documents)
        else:
            super().add(documents)

        return documents

    def index(self):
        """
        Builds an embeddings index for previously batched documents.
        """

        if self.cluster:
            self.cluster.index()
        else:
            super().index()

    def upsert(self):
        """
        Runs an embeddings upsert operation for previously batched documents.
        """

        if self.cluster:
            self.cluster.upsert()
        else:
            super().upsert()

    def delete(self, ids):
        """
        Deletes from an embeddings index. Returns list of ids deleted.

        Args:
            ids: list of ids to delete

        Returns:
            ids deleted
        """

        if self.cluster:
            return self.cluster.delete(ids)

        return super().delete(ids)

    def reindex(self, config, function=None):
        """
        Recreates this embeddings index using config. This method only works if document content storage is enabled.

        Args:
            config: new config
            function: optional function to prepare content for indexing
        """

        if self.cluster:
            self.cluster.reindex(config, function)
        else:
            super().reindex(config, function)

    def count(self):
        """
        Total number of elements in this embeddings index.

        Returns:
            number of elements in embeddings index
        """

        if self.cluster:
            return self.cluster.count()

        return super().count()

    def limit(self, limit):
        """
        Parses the number of results to return from the request. Allows range of 1-250, with a default of 10.

        Args:
            limit: limit parameter

        Returns:
            bounded limit
        """

        # Return between 1 and 250 results, defaults to 10
        return max(1, min(250, int(limit) if limit else 10))

    def weights(self, weights):
        """
        Parses the weights parameter from the request.

        Args:
            weights: weights parameter

        Returns:
            weights
        """

        return float(weights) if weights else weights



"""
Defines API paths for similarity endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.post("/similarity")
def similarity(query: str = Body(...), texts: List[str] = Body(...)):
    """
    Computes the similarity between query and list of text. Returns a list of
    {id: value, score: value} sorted by highest score, where id is the index
    in texts.

    Args:
        query: query text
        texts: list of text

    Returns:
        list of {id: value, score: value}
    """

    return application.get().similarity(query, texts)


@router.post("/batchsimilarity")
def batchsimilarity(queries: List[str] = Body(...), texts: List[str] = Body(...)):
    """
    Computes the similarity between list of queries and list of text. Returns a list
    of {id: value, score: value} sorted by highest score per query, where id is the
    index in texts.

    Args:
        queries: queries text
        texts: list of text

    Returns:
        list of {id: value, score: value} per query
    """

    return application.get().batchsimilarity(queries, texts)



"""
Hugging Face module
"""

from ...models import Models, PoolingFactory

from ..base import Vectors


class HFVectors(Vectors):
    """
    Builds vectors using the Hugging Face transformers library.
    """

    @staticmethod
    def ismethod(method):
        """
        Checks if this method uses local transformers-based models.

        Args:
            method: input method

        Returns:
            True if this is a local transformers-based model, False otherwise
        """

        return method in ("transformers", "pooling", "clspooling", "meanpooling")

    def loadmodel(self, path):
        # Build embeddings with transformers pooling
        return PoolingFactory.create(
            {
                "method": self.config.get("method"),
                "path": path,
                "device": Models.deviceid(self.config.get("gpu", True)),
                "tokenizer": self.config.get("tokenizer"),
                "maxlength": self.config.get("maxlength"),
                "modelargs": self.config.get("vectors", {}),
            }
        )

    def encode(self, data, category=None):
        # Encode data using vectors model
        return self.model.encode(data, batch=self.encodebatch, category=category)



"""
API imports
"""

# Conditional import
try:
    from .authorization import Authorization
    from .application import app, start
    from .base import API
    from .cluster import Cluster
    from .extension import Extension
    from .factory import APIFactory
    from .responses import *
    from .routers import *
    from .route import EncodingAPIRoute
except ImportError as missing:
    # pylint: disable=W0707
    raise ImportError('API is not available - install "api" extra to enable') from missing



"""
ANN (Approximate Nearest Neighbor) module
"""

import datetime
import platform

from ..version import __version__


class ANN:
    """
    Base class for ANN instances. This class builds vector indexes to support similarity search.
    The built-in ANN backends store ids and vectors. Content storage is supported via database instances.
    """

    def __init__(self, config):
        """
        Creates a new ANN.

        Args:
            config: index configuration parameters
        """

        # ANN index
        self.backend = None

        # ANN configuration
        self.config = config

    def load(self, path):
        """
        Loads an ANN at path.

        Args:
            path: path to load ann index
        """

        raise NotImplementedError

    def index(self, embeddings):
        """
        Builds an ANN index.

        Args:
            embeddings: embeddings array
        """

        raise NotImplementedError

    def append(self, embeddings):
        """
        Append elements to an existing index.

        Args:
            embeddings: embeddings array
        """

        raise NotImplementedError

    def delete(self, ids):
        """
        Deletes elements from existing index.

        Args:
            ids: ids to delete
        """

        raise NotImplementedError

    def search(self, queries, limit):
        """
        Searches ANN index for query. Returns topn results.

        Args:
            queries: queries array
            limit: maximum results

        Returns:
            query results
        """

        raise NotImplementedError

    def count(self):
        """
        Number of elements in the ANN index.

        Returns:
            count
        """

        raise NotImplementedError

    def save(self, path):
        """
        Saves an ANN index at path.

        Args:
            path: path to save ann index
        """

        raise NotImplementedError

    def close(self):
        """
        Closes this ANN.
        """

        self.backend = None

    def setting(self, name, default=None):
        """
        Looks up backend specific setting.

        Args:
            name: setting name
            default: default value when setting not found

        Returns:
            setting value
        """

        # Get the backend-specific config object
        backend = self.config.get(self.config["backend"])

        # Get setting value, set default value if not found
        setting = backend.get(name) if backend else None
        return setting if setting else default

    def metadata(self, settings=None):
        """
        Adds index build metadata.

        Args:
            settings: index build settings
        """

        # ISO 8601 timestamp
        create = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

        # Set build metadata if this is not an update
        if settings:
            self.config["build"] = {
                "create": create,
                "python": platform.python_version(),
                "settings": settings,
                "system": f"{platform.system()} ({platform.machine()})",
                "txtai": __version__,
            }

        # Set last update date
        self.config["update"] = create



"""
Archive module
"""

import os

from tempfile import TemporaryDirectory

from .tar import Tar
from .zip import Zip


class Archive:
    """
    Base class for archive instances.
    """

    def __init__(self, directory=None):
        """
        Creates a new archive instance.

        Args:
            directory: directory to use as working directory, defaults to a temporary directory
        """

        self.directory = directory

    def isarchive(self, path):
        """
        Checks if path is an archive file based on the extension.

        Args:
            path: path to check

        Returns:
            True if the path ends with an archive extension, False otherwise
        """

        return path and any(path.lower().endswith(extension) for extension in [".tar.bz2", ".tar.gz", ".tar.xz", ".zip"])

    def path(self):
        """
        Gets the current working directory for this archive instance.

        Returns:
            archive working directory
        """

        # Default to a temporary directory. All files created in this directory will be deleted
        # when this archive instance goes out of scope.
        if not self.directory:
            # pylint: disable=R1732
            self.directory = TemporaryDirectory()

        return self.directory.name if isinstance(self.directory, TemporaryDirectory) else self.directory

    def load(self, path, compression=None):
        """
        Extracts file at path to archive working directory.

        Args:
            path: path to archive file
            compression: compression format, infers from path if not provided
        """

        # Unpack compressed file
        compress = self.create(path, compression)
        compress.unpack(path, self.path())

    def save(self, path, compression=None):
        """
        Archives files in archive working directory to file at path.

        Args:
            path: path to archive file
            compression: compression format, infers from path if not provided
        """

        # Create output directory, if necessary
        output = os.path.dirname(path)
        if output:
            os.makedirs(output, exist_ok=True)

        # Pack into compressed file
        compress = self.create(path, compression)
        compress.pack(self.path(), path)

    def create(self, path, compression):
        """
        Method to construct a Compress instance.

        Args:
            path: file path
            compression: compression format, infers using file extension if not provided

        Returns:
            Compress
        """

        # Infer compression format from path if not provided
        compression = compression if compression else path.lower().split(".")[-1]

        # Create compression instance
        return Zip() if compression == "zip" else Tar()



"""
Authorization module
"""

import hashlib
import os

from fastapi import Header, HTTPException


class Authorization:
    """
    Basic token authorization.
    """

    def __init__(self, token=None):
        """
        Creates a new Authorization instance.

        Args:
            token: SHA-256 hash of token to check
        """

        self.token = token if token else os.environ.get("TOKEN")

    def __call__(self, authorization: str = Header(default=None)):
        """
        Validates authorization header is present and equal to current token.

        Args:
            authorization: authorization header
        """

        if not authorization or self.token != self.digest(authorization):
            raise HTTPException(status_code=401, detail="Invalid Authorization Token")

    def digest(self, authorization):
        """
        Computes a SHA-256 hash for input authorization token.

        Args:
            authorization: authorization header

        Returns:
            SHA-256 hash of authorization token
        """

        # Replace Bearer prefix
        prefix = "Bearer "
        token = authorization[len(prefix) :] if authorization.startswith(prefix) else authorization

        # Compute SHA-256 hash
        return hashlib.sha256(token.encode("utf-8")).hexdigest()



"""
Late module
"""

import json

import numpy as np
import torch

from huggingface_hub.errors import HFValidationError
from safetensors import safe_open
from torch import nn
from transformers.utils import cached_file

from .base import Pooling
from .muvera import Muvera


class LatePooling(Pooling):
    """
    Builds late pooled vectors using outputs from a transformers model.
    """

    def __init__(self, path, device, tokenizer=None, maxlength=None, modelargs=None):
        # Check if fixed dimensional encoder is enabled
        modelargs = modelargs.copy() if modelargs else {}
        muvera = modelargs.pop("muvera", {})
        self.encoder = Muvera(**muvera) if muvera is not None else None

        # Call parent initialization
        super().__init__(path, device, tokenizer, maxlength, modelargs)

        # Get linear weights path
        config = self.load(path, "1_Dense/config.json")
        if config:
            # PyLate weights format
            name = "1_Dense/model.safetensors"
        else:
            # Stanford weights format
            name = "model.safetensors"

        # Read model settings
        self.qprefix, self.qlength, self.dprefix, self.dlength = self.settings(path, config)

        # Load linear layer
        path = cached_file(path_or_repo_id=path, filename=name)
        with safe_open(filename=path, framework="pt") as f:
            weights = f.get_tensor("linear.weight")

            # Load weights into linear layer
            self.linear = nn.Linear(weights.shape[1], weights.shape[0], bias=False, device=self.device, dtype=weights.dtype)
            with torch.no_grad():
                self.linear.weight.copy_(weights)

    def forward(self, **inputs):
        """
        Runs late pooling on token embeddings.

        Args:
            inputs: model inputs

        Returns:
            Late pooled embeddings using output token embeddings (i.e. last hidden state)
        """

        # Run through transformers model
        tokens = super().forward(**inputs)

        # Run through final linear layer and return
        return self.linear(tokens)

    def preencode(self, documents, category):
        """
        Apply prefixes and lengths to data.

        Args:
            documents: list of documents used to build embeddings
            category: embeddings category (query or data)
        """

        results = []

        # Apply prefix
        for text in documents:
            prefix = self.qprefix if category == "query" else self.dprefix
            if prefix:
                text = f"{prefix}{text}"

            results.append(text)

        # Set maxlength
        maxlength = self.qlength if category == "query" else self.dlength
        if maxlength:
            self.maxlength = maxlength

        return results

    def postencode(self, results, category):
        """
        Normalizes and pads results.

        Args:
            results: input results

        Returns:
            normalized results with padding
        """

        length = 0
        for vectors in results:
            # Get max length
            if vectors.shape[0] > length:
                length = vectors.shape[0]

            # Normalize vectors
            vectors /= np.linalg.norm(vectors, axis=1)[:, np.newaxis]

        # Pad values
        data = []
        for vectors in results:
            data.append(np.pad(vectors, [(0, length - vectors.shape[0]), (0, 0)]))

        # Build NumPy array
        data = np.asarray(data)

        # Apply fixed dimesional encoder, if necessary
        return self.encoder(data, category) if self.encoder else data

    def settings(self, path, config):
        """
        Reads model settings.

        Args:
            path: model path
            config: PyLate model format if provided, otherwise read from Stanford format
        """

        if config:
            # PyLate format
            config = self.load(path, "config_sentence_transformers.json")
            params = ["query_prefix", "query_length", "document_prefix", "document_length"]
        else:
            # Stanford format
            config = self.load(path, "artifact.metadata")
            params = ["query_token_id", "query_maxlen", "doc_token_id", "doc_maxlen"]

        return [config.get(p) for p in params]

    def load(self, path, name):
        """
        Loads a JSON config file from the Hugging Face Hub.

        Args:
            path: model path
            name: file to load

        Returns:
            config
        """

        # Download file and parse JSON
        config = None
        try:
            path = cached_file(path_or_repo_id=path, filename=name)
            if path:
                with open(path, encoding="utf-8") as f:
                    config = json.load(f)

        # Ignore this error - invalid repo or directory
        except (HFValidationError, OSError):
            pass

        return config



"""
Dense vectors imports
"""

from .external import External
from .factory import VectorsFactory
from .huggingface import HFVectors
from .litellm import LiteLLM
from .llama import LlamaCpp
from .m2v import Model2Vec
from .sbert import STVectors
from .words import WordVectors



"""
Agent imports
"""

# Conditional import
try:
    from .base import Agent
    from .factory import ProcessFactory
    from .model import PipelineModel
    from .tool import *
except ImportError:
    from .placeholder import Agent



"""
LiteLLM module
"""

from transformers.utils import cached_file

# Conditional import
try:
    import litellm as api

    LITELLM = True
except ImportError:
    LITELLM = False

from .generation import Generation


class LiteLLM(Generation):
    """
    LiteLLM generative model.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a LiteLLM model.

        Args:
            path: input path

        Returns:
            True if this is a LiteLLM model, False otherwise
        """

        # pylint: disable=W0702
        if isinstance(path, str) and LITELLM:
            debug = api.suppress_debug_info
            try:
                # Suppress debug messages for this test
                api.suppress_debug_info = True
                return api.get_llm_provider(path) and not LiteLLM.ishub(path)
            except:
                return False
            finally:
                # Restore debug info value to original value
                api.suppress_debug_info = debug

        return False

    @staticmethod
    def ishub(path):
        """
        Checks if path is available on the HF Hub.

        Args:
            input path

        Returns:
            True if this is a model on the HF Hub
        """

        # pylint: disable=W0702
        try:
            return cached_file(path_or_repo_id=path, filename="config.json") is not None if "/" in path else False
        except:
            return False

    def __init__(self, path, template=None, **kwargs):
        super().__init__(path, template, **kwargs)

        if not LITELLM:
            raise ImportError('LiteLLM is not available - install "pipeline" extra to enable')

        # Ignore common pipeline parameters
        self.kwargs = {k: v for k, v in self.kwargs.items() if k not in ["quantize", "gpu", "model", "task"]}

    def stream(self, texts, maxlength, stream, stop, **kwargs):
        for text in texts:
            # LLM API call
            result = api.completion(
                model=self.path,
                messages=[{"content": text, "role": "prompt"}] if isinstance(text, str) else text,
                max_tokens=maxlength,
                stream=stream,
                stop=stop,
                **{**self.kwargs, **kwargs}
            )

            # Stream response
            yield from self.response(result if stream else [result])



"""
Serialize imports
"""

from .base import Serialize
from .errors import SerializeError
from .factory import SerializeFactory
from .messagepack import MessagePack
from .pickle import Pickle
from .serializer import Serializer



"""
IVFSparse module
"""

import math
import os

from multiprocessing.pool import ThreadPool

import numpy as np

# Conditional import
try:
    from scipy.sparse import csr_matrix, vstack
    from scipy.sparse.linalg import norm
    from sklearn.cluster import MiniBatchKMeans
    from sklearn.metrics import pairwise_distances_argmin_min
    from sklearn.utils.extmath import safe_sparse_dot

    IVFSPARSE = True
except ImportError:
    IVFSPARSE = False

from ...serialize import SerializeFactory
from ...util import SparseArray
from ..base import ANN


class IVFSparse(ANN):
    """
    Inverted file (IVF) index with flat vector file storage and sparse array support.

    IVFSparse builds an IVF index and enables approximate nearest neighbor (ANN) search.

    This index is modeled after Faiss and supports many of the same parameters.

    See this link for more: https://github.com/facebookresearch/faiss/wiki/Faster-search
    """

    def __init__(self, config):
        super().__init__(config)

        if not IVFSPARSE:
            raise ImportError('IVFSparse is not available - install "ann" extra to enable')

        # Cluster centroids, if computed
        self.centroids = None

        # Cluster id mapping
        self.ids = None

        # Cluster data blocks - can be a single block with no computed centroids
        self.blocks = None

        # Deleted ids
        self.deletes = None

    def index(self, embeddings):
        # Compute model training size
        train, sample = embeddings, self.setting("sample")
        if sample:
            # Get sample for training
            rng = np.random.default_rng(0)
            indices = sorted(rng.choice(train.shape[0], int(sample * train.shape[0]), replace=False, shuffle=False))
            train = train[indices]

        # Get number of clusters. Note that final number of clusters could be lower due to filtering duplicate centroids
        # and pruning of small clusters
        clusters = self.nlist(embeddings.shape[0], train.shape[0])

        # Build cluster centroids if approximate search is enabled
        # A single cluster performs exact search
        self.centroids = self.build(train, clusters) if clusters > 1 else None

        # Sort into clusters
        ids = self.aggregate(embeddings)

        # Prune small clusters (less than minpoints parameter) and rebuild
        indices = sorted(k for k, v in ids.items() if len(v) >= self.minpoints())
        if len(indices) > 0 and len(ids) > 1 and len(indices) != len(ids.keys()):
            self.centroids = self.centroids[indices]
            ids = self.aggregate(embeddings)

        # Sort clusters by id
        self.ids = dict(sorted(ids.items(), key=lambda x: x[0]))

        # Create cluster data blocks
        self.blocks = {k: embeddings[v] for k, v in self.ids.items()}

        # Calculate block max summary vectors and use as centroids
        self.centroids = vstack([csr_matrix(x.max(axis=0)) for x in self.blocks.values()]) if self.centroids is not None else None

        # Initialize deletes
        self.deletes = []

        # Add id offset and index build metadata
        self.config["offset"] = embeddings.shape[0]
        self.metadata({"clusters": len(self.blocks)})

    def append(self, embeddings):
        # Get offset
        offset = self.size()

        # Sort into clusters and merge
        for cluster, ids in self.aggregate(embeddings).items():
            # Add new ids
            self.ids[cluster].extend([x + offset for x in ids])

            # Add new data
            self.blocks[cluster] = vstack([self.blocks[cluster], embeddings[ids]])

        # Update id offset and index metadata
        self.config["offset"] += embeddings.shape[0]
        self.metadata()

    def delete(self, ids):
        # Set index ids as deleted
        self.deletes.extend(ids)

    def search(self, queries, limit):
        results = []

        # Calculate number of threads using a thread batch size of 32
        threads = queries.shape[0] // 32
        threads = min(max(threads, 1), os.cpu_count())

        # Approximate search
        blockids = self.topn(queries, self.centroids, self.nprobe())[0] if self.centroids is not None else None

        # This method is able to run as multiple threads due to a number of numpy/scipy method calls that drop the GIL.
        results = []
        with ThreadPool(threads) as pool:
            for result in pool.starmap(self.scan, [(x, limit, blockids[i] if blockids is not None else None) for i, x in enumerate(queries)]):
                results.append(result)

        return results

    def count(self):
        return self.size() - len(self.deletes)

    def load(self, path):
        # Create streaming serializer and limit read size to a byte at a time to ensure
        # only msgpack data is consumed
        serializer = SerializeFactory.create("msgpack", streaming=True, read_size=1)

        with open(path, "rb") as f:
            # Read header
            unpacker = serializer.loadstream(f)
            header = next(unpacker)

            # Read cluster centroids, if available
            self.centroids = SparseArray().load(f) if header["centroids"] else None

            # Read cluster ids
            self.ids = dict(next(unpacker))

            # Read cluster data blocks
            self.blocks = {}
            for key in self.ids:
                self.blocks[key] = SparseArray().load(f)

            # Read deletes
            self.deletes = next(unpacker)

    def save(self, path):
        # IVFSparse storage format:
        #    - header msgpack
        #    - centroids sparse array (optional based on header parameters)
        #    - cluster ids msgpack
        #    - cluster data blocks list of sparse arrays
        #    - deletes msgpack

        # Create message pack serializer
        serializer = SerializeFactory.create("msgpack")

        with open(path, "wb") as f:
            # Write header
            serializer.savestream({"centroids": self.centroids is not None, "count": self.count(), "blocks": len(self.blocks)}, f)

            # Write cluster centroids, if available
            if self.centroids is not None:
                SparseArray().save(f, self.centroids)

            # Write cluster id mapping
            serializer.savestream(list(self.ids.items()), f)

            # Write cluster data blocks
            for block in self.blocks.values():
                SparseArray().save(f, block)

            # Write deletes
            serializer.savestream(self.deletes, f)

    def build(self, train, clusters):
        """
        Builds a k-means cluster to calculate centroid points for aggregating data blocks.

        Args:
            train: training data
            clusters: number of clusters to create

        Returns:
            cluster centroids
        """

        # Select top n most important features that contribute to L2 vector norm
        indices = np.argsort(-norm(train, axis=0))[: self.setting("nfeatures", 25)]
        data = train[:, indices]
        data = train

        # Cluster data using k-means
        kmeans = MiniBatchKMeans(n_clusters=clusters, random_state=0, n_init=5).fit(data)

        # Find closest points to each cluster center and use those as centroids
        indices, _ = pairwise_distances_argmin_min(kmeans.cluster_centers_, data, metric="l2")

        # Filter out duplicate centroids and return cluster centroids
        return train[np.unique(indices)]

    def aggregate(self, data):
        """
        Aggregates input data array into clusters. This method sorts each data element into the
        cluster with the highest L2 similarity centroid.

        Args:
            data: input data

        Returns:
            {cluster, ids}
        """

        # Exact search when only a single cluster
        if self.centroids is None:
            return {0: list(range(data.shape[0]))}

        # Map data to closest centroids
        indices, _ = pairwise_distances_argmin_min(data, self.centroids, metric="l2")

        # Sort into clusters
        ids = {}
        for x, cluster in enumerate(indices.tolist()):
            if cluster not in ids:
                ids[cluster] = []

            # Save id
            ids[cluster].append(x)

        return ids

    def topn(self, queries, data, limit, deletes=None):
        """
        Gets the top n most similar data elements for query.

        Args:
            queries: queries array
            data: data array
            limit: top n
            deletes: optional list of deletes to filter from results

        Returns:
            list of matching (indices, scores)
        """

        # Dot product similarity
        scores = safe_sparse_dot(queries, data.T, dense_output=True)

        # Clear deletes
        if deletes is not None:
            scores[:, deletes] = 0

        # Get top n matching indices and scores
        indices = np.argpartition(-scores, limit if limit < scores.shape[0] else scores.shape[0] - 1)[:, :limit]
        scores = np.take_along_axis(scores, indices, axis=1)

        return indices, scores

    def scan(self, query, limit, blockids):
        """
        Scans a list of blocks for top n ids that match query.

        Args:
            query: input query
            limit top n
            blockids: block ids to scan

        Returns:
            list of (id, scores)
        """

        if self.centroids is not None:
            # Stack into single ids list
            ids = np.concatenate([self.ids[x] for x in blockids if x in self.ids])

            # Stack data rows
            data = vstack([self.blocks[x] for x in blockids if x in self.blocks])
        else:
            # Exact search
            ids, data = np.array(self.ids[0]), self.blocks[0]

        # Get deletes
        deletes = np.argwhere(np.isin(ids, self.deletes)).ravel()

        # Calculate similarity
        indices, scores = self.topn(query, data, limit, deletes)
        indices, scores = indices[0], scores[0]

        # Map data ids and return
        return list(zip(ids[indices].tolist(), scores.tolist()))

    def nlist(self, count, train):
        """
        Calculates the number of clusters for this IVFSparse index. Note that the final number of clusters
        could be lower as duplicate cluster centroids are filtered out.

        Args:
            count: initial dataset size
            train: number of rows used to train

        Returns:
            number of clusters
        """

        # Get data size
        default = 1 if count <= 5000 else self.cells(train)

        # Number of clusters to create
        return self.setting("nlist", default)

    def nprobe(self):
        """
        Gets or derives the nprobe search parameter.

        Returns:
            nprobe setting
        """

        # Get size of embeddings index
        size = self.size()

        default = 6 if size <= 5000 else self.cells(size) // 16
        return self.setting("nprobe", default)

    def cells(self, count):
        """
        Calculates the number of IVF cells for an IVFSparse index.

        Args:
            count: number of rows

        Returns:
            number of IVF cells
        """

        # Calculate number of IVF cells where x = min(4 * sqrt(count), count / minpoints)
        return max(min(round(4 * math.sqrt(count)), int(count / self.minpoints())), 1)

    def size(self):
        """
        Gets the total size of this index including deletes.

        Returns:
            size
        """

        return sum(len(x) for x in self.ids.values())

    def minpoints(self):
        """
        Gets the minimum number of points per cluster.

        Returns:
            minimum points per cluster
        """

        # Minimum number of points per cluster
        # Match faiss default that requires at least 39 points per clusters
        return self.setting("minpoints", 39)



"""
Defines API paths for embeddings endpoints.
"""

from io import BytesIO
from typing import List, Optional

import PIL

from fastapi import APIRouter, Body, File, Form, HTTPException, Request, UploadFile
from fastapi.encoders import jsonable_encoder

from .. import application
from ..responses import ResponseFactory
from ..route import EncodingAPIRoute

from ...app import ReadOnlyError
from ...graph import Graph

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/search")
def search(query: str, request: Request):
    """
    Finds documents most similar to the input query. This method will run either an index search
    or an index + database search depending on if a database is available.

    Args:
        query: input query
        request: FastAPI request

    Returns:
        list of {id: value, score: value} for index search, list of dict for an index + database search
    """

    # Execute search
    results = application.get().search(query, request=request)

    # Encode using standard FastAPI encoder but skip certain classes
    results = jsonable_encoder(
        results, custom_encoder={bytes: lambda x: x, BytesIO: lambda x: x, PIL.Image.Image: lambda x: x, Graph: lambda x: x.savedict()}
    )

    # Return raw response to prevent duplicate encoding
    response = ResponseFactory.create(request)
    return response(results)


# pylint: disable=W0621
@router.post("/batchsearch")
def batchsearch(
    request: Request,
    queries: List[str] = Body(...),
    limit: int = Body(default=None),
    weights: float = Body(default=None),
    index: str = Body(default=None),
    parameters: List[dict] = Body(default=None),
    graph: bool = Body(default=False),
):
    """
    Finds documents most similar to the input queries. This method will run either an index search
    or an index + database search depending on if a database is available.

    Args:
        queries: input queries
        limit: maximum results
        weights: hybrid score weights, if applicable
        index: index name, if applicable
        parameters: list of dicts of named parameters to bind to placeholders
        graph: return graph results if True

    Returns:
        list of {id: value, score: value} per query for index search, list of dict per query for an index + database search
    """

    # Execute search
    results = application.get().batchsearch(queries, limit, weights, index, parameters, graph)

    # Encode using standard FastAPI encoder but skip certain classes
    results = jsonable_encoder(
        results, custom_encoder={bytes: lambda x: x, BytesIO: lambda x: x, PIL.Image.Image: lambda x: x, Graph: lambda x: x.savedict()}
    )

    # Return raw response to prevent duplicate encoding
    response = ResponseFactory.create(request)
    return response(results)


@router.post("/add")
def add(documents: List[dict] = Body(...)):
    """
    Adds a batch of documents for indexing.

    Args:
        documents: list of {id: value, text: value, tags: value}
    """

    try:
        application.get().add(documents)
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.post("/addobject")
def addobject(data: List[bytes] = File(), uid: List[str] = Form(default=None), field: str = Form(default=None)):
    """
    Adds a batch of binary documents for indexing.

    Args:
        data: list of binary objects
        uid: list of corresponding ids
        field: optional object field name
    """

    if uid and len(data) != len(uid):
        raise HTTPException(status_code=422, detail="Length of data and document lists must match")

    try:
        # Add objects
        application.get().addobject(data, uid, field)
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.post("/addimage")
def addimage(data: List[UploadFile] = File(), uid: List[str] = Form(), field: str = Form(default=None)):
    """
    Adds a batch of images for indexing.

    Args:
        data: list of images
        uid: list of corresponding ids
        field: optional object field name
    """

    if uid and len(data) != len(uid):
        raise HTTPException(status_code=422, detail="Length of data and uid lists must match")

    try:
        # Add images
        application.get().addobject([PIL.Image.open(content.file) for content in data], uid, field)
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.get("/index")
def index():
    """
    Builds an embeddings index for previously batched documents.
    """

    try:
        application.get().index()
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.get("/upsert")
def upsert():
    """
    Runs an embeddings upsert operation for previously batched documents.
    """

    try:
        application.get().upsert()
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.post("/delete")
def delete(ids: List = Body(...)):
    """
    Deletes from an embeddings index. Returns list of ids deleted.

    Args:
        ids: list of ids to delete

    Returns:
        ids deleted
    """

    try:
        return application.get().delete(ids)
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.post("/reindex")
def reindex(config: dict = Body(...), function: str = Body(default=None)):
    """
    Recreates this embeddings index using config. This method only works if document content storage is enabled.

    Args:
        config: new config
        function: optional function to prepare content for indexing
    """

    try:
        application.get().reindex(config, function)
    except ReadOnlyError as e:
        raise HTTPException(status_code=403, detail=e.args[0]) from e


@router.get("/count")
def count():
    """
    Total number of elements in this embeddings index.

    Returns:
        number of elements in embeddings index
    """

    return application.get().count()


@router.post("/explain")
def explain(query: str = Body(...), texts: List[str] = Body(default=None), limit: int = Body(default=None)):
    """
    Explains the importance of each input token in text for a query.

    Args:
        query: query text
        texts: list of text

    Returns:
        list of dict where a higher scores represents higher importance relative to the query
    """

    return application.get().explain(query, texts, limit)


@router.post("/batchexplain")
def batchexplain(queries: List[str] = Body(...), texts: List[str] = Body(default=None), limit: int = Body(default=None)):
    """
    Explains the importance of each input token in text for a query.

    Args:
        query: query text
        texts: list of text

    Returns:
        list of dict where a higher scores represents higher importance relative to the query
    """

    return application.get().batchexplain(queries, texts, limit)


@router.get("/transform")
def transform(text: str, category: Optional[str] = None, index: Optional[str] = None):
    """
    Transforms text into an embeddings array.

    Args:
        text: input text
        category: category for instruction-based embeddings
        index: index name, if applicable

    Returns:
        embeddings array
    """

    return application.get().transform(text, category, index)


@router.post("/batchtransform")
def batchtransform(texts: List[str] = Body(...), category: Optional[str] = None, index: Optional[str] = None):
    """
    Transforms list of text into embeddings arrays.

    Args:
        texts: list of text
        category: category for instruction-based embeddings
        index: index name, if applicable

    Returns:
        embeddings arrays
    """

    return application.get().batchtransform(texts, category, index)



"""
Tokenizer module
"""

import re
import string

import regex

from ..base import Pipeline


class Tokenizer(Pipeline):
    """
    Tokenizes text into tokens using one of the following methods.

      1. Backwards compatible tokenization that only accepts alphanumeric tokens from the Latin alphabet.

      2. Split using word boundary rules from the Unicode Text Segmentation algorithm (see Unicode Standard Annex #29).
         This is similar to the standard tokenizer in Apache Lucene and works well for most languages.
    """

    # fmt: off
    # English Stop Word List (Standard stop words used by Apache Lucene)
    STOP_WORDS = {"a", "an", "and", "are", "as", "at", "be", "but", "by", "for", "if", "in", "into", "is", "it",
                  "no", "not", "of", "on", "or", "such", "that", "the", "their", "then", "there", "these",
                  "they", "this", "to", "was", "will", "with"}
    # fmt: on

    @staticmethod
    def tokenize(text, lowercase=True, emoji=True, alphanum=True, stopwords=True):
        """
        Tokenizes text into a list of tokens. The default backwards compatible parameters filter out English stop words and only
        accept alphanumeric tokens.

        Args:
            text: input text
            lowercase: lower cases all tokens if True, defaults to True
            emoji: tokenize emoji in text if True, defaults to True
            alphanum: requires 2+ character alphanumeric tokens if True, defaults to True
            stopwords: removes provided stop words if a list, removes default English stop words if True, defaults to True

        Returns:
            list of tokens
        """

        # Create a tokenizer with backwards compatible settings
        return Tokenizer(lowercase, emoji, alphanum, stopwords)(text)

    def __init__(self, lowercase=True, emoji=True, alphanum=False, stopwords=False):
        """
        Creates a new tokenizer. The default parameters segment text per Unicode Standard Annex #29.

        Args:
            lowercase: lower cases all tokens if True, defaults to True
            emoji: tokenize emoji in text if True, defaults to True
            alphanum: requires 2+ character alphanumeric tokens if True, defaults to False
            stopwords: removes provided stop words if a list, removes default English stop words if True, defaults to False
        """

        # Lowercase
        self.lowercase = lowercase

        # Text segmentation
        self.alphanum, self.segment = None, None
        if alphanum:
            # Alphanumeric regex that accepts tokens that meet following rules:
            #  - Strings to be at least 2 characters long AND
            #  - At least 1 non-trailing alpha character in string
            # Note: The standard Python re module is much faster than regex for this expression
            self.alphanum = re.compile(r"^\d*[a-z][\-.0-9:_a-z]{1,}$")
        else:
            # Text segmentation per Unicode Standard Annex #29
            pattern = r"\w\p{Extended_Pictographic}\p{WB:RegionalIndicator}" if emoji else r"\w"
            self.segment = regex.compile(rf"[{pattern}](?:\B\S)*", flags=regex.WORD)

        # Stop words
        self.stopwords = stopwords if isinstance(stopwords, list) else Tokenizer.STOP_WORDS if stopwords else False

    def __call__(self, text):
        """
        Tokenizes text into a list of tokens.

        Args:
            text: input text

        Returns:
            list of tokens
        """

        # Check for None and skip processing
        if text is None:
            return None

        # Lowercase
        text = text.lower() if self.lowercase else text

        if self.alphanum:
            # Text segmentation using standard split
            tokens = [token.strip(string.punctuation) for token in text.split()]

            # Filter on alphanumeric strings.
            tokens = [token for token in tokens if re.match(self.alphanum, token)]
        else:
            # Text segmentation per Unicode Standard Annex #29
            tokens = regex.findall(self.segment, text)

        # Stop words
        if self.stopwords:
            tokens = [token for token in tokens if token not in self.stopwords]

        return tokens



"""
Factory module
"""

from .messagepack import MessagePack
from .pickle import Pickle


class SerializeFactory:
    """
    Methods to create data serializers.
    """

    @staticmethod
    def create(method=None, **kwargs):
        """
        Creates a new Serialize instance.

        Args:
            method: serialization method
            kwargs: additional keyword arguments to pass to serialize instance
        """

        # Pickle serialization
        if method == "pickle":
            return Pickle(**kwargs)

        # Default serialization
        return MessagePack(**kwargs)



"""
Defines API paths for summary endpoints.
"""

from typing import List, Optional

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/summary")
def summary(text: str, minlength: Optional[int] = None, maxlength: Optional[int] = None):
    """
    Runs a summarization model against a block of text.

    Args:
        text: text to summarize
        minlength: minimum length for summary
        maxlength: maximum length for summary

    Returns:
        summary text
    """

    return application.get().pipeline("summary", (text, minlength, maxlength))


@router.post("/batchsummary")
def batchsummary(texts: List[str] = Body(...), minlength: Optional[int] = Body(default=None), maxlength: Optional[int] = Body(default=None)):
    """
    Runs a summarization model against a block of text.

    Args:
        texts: list of text to summarize
        minlength: minimum length for summary
        maxlength: maximum length for summary

    Returns:
        list of summary text
    """

    return application.get().pipeline("summary", (texts, minlength, maxlength))



"""
CLS module
"""

from .base import Pooling


class ClsPooling(Pooling):
    """
    Builds CLS pooled vectors using outputs from a transformers model.
    """

    def forward(self, **inputs):
        """
        Runs CLS pooling on token embeddings.

        Args:
            inputs: model inputs

        Returns:
            CLS pooled embeddings using output token embeddings (i.e. last hidden state)
        """

        # Run through transformers model
        tokens = super().forward(**inputs)

        # CLS token pooling
        return tokens[:, 0]



"""
Console module
"""

import os
import shlex

from cmd import Cmd

# Conditional import
try:
    from rich import box
    from rich.console import Console as RichConsole
    from rich.table import Table

    RICH = True
except ImportError:
    RICH = False

from txtai.app import Application
from txtai.embeddings import Embeddings


class Console(Cmd):
    """
    txtai console.
    """

    def __init__(self, path=None):
        """
        Creates a new command line console.

        Args:
            path: path to initial configuration, if any
        """

        super().__init__()

        if not RICH:
            raise ImportError('Console is not available - install "console" extra to enable')

        self.prompt = ">>> "

        # Rich console
        self.console = RichConsole()

        # App parameters
        self.app = None
        self.path = path

        # Parameters
        self.vhighlight = None
        self.vlimit = None

    def preloop(self):
        """
        Loads initial configuration.
        """

        self.console.print("txtai console", style="#03a9f4")

        # Load default path
        if self.path:
            self.load(self.path)

    def default(self, line):
        """
        Default event loop.

        Args:
            line: command line
        """

        # pylint: disable=W0703
        try:
            command = line.lower()
            if command.startswith(".config"):
                self.config()
            elif command.startswith(".highlight"):
                self.highlight(command)
            elif command.startswith(".limit"):
                self.limit(command)
            elif command.startswith(".load"):
                command = self.split(line)
                self.path = command[1]
                self.load(self.path)
            elif command.startswith(".workflow"):
                self.workflow(line)
            else:
                # Search is default action
                self.search(line)
        except Exception:
            self.console.print_exception()

    def config(self):
        """
        Processes .config command.
        """

        self.console.print(self.app.config)

    def highlight(self, command):
        """
        Processes .highlight command.

        Args:
            command: command line
        """

        _, action = self.split(command, "#ffff00")
        self.vhighlight = action
        self.console.print(f"Set highlight to {self.vhighlight}")

    def limit(self, command):
        """
        Processes .limit command.

        Args:
            command: command line
        """

        _, action = self.split(command, 10)
        self.vlimit = int(action)
        self.console.print(f"Set limit to {self.vlimit}")

    def load(self, path):
        """
        Processes .load command.

        Args:
            path: path to configuration
        """

        if self.isyaml(path):
            self.console.print(f"Loading application {path}")
            self.app = Application(path)
        else:
            self.console.print(f"Loading index {path}")

            # Load embeddings index
            self.app = Embeddings()
            self.app.load(path)

    def search(self, query):
        """
        Runs a search query.

        Args:
            query: query to run
        """

        if self.vhighlight:
            results = self.app.explain(query, limit=self.vlimit)
        else:
            results = self.app.search(query, limit=self.vlimit)

        columns, table = {}, Table(box=box.SQUARE, style="#03a9f4")

        # Build column list
        result = results[0]
        if isinstance(result, tuple):
            columns = dict.fromkeys(["id", "score"])
        else:
            columns = dict(result)

        # Add columns to table
        columns = list(x for x in columns if x != "tokens")
        for column in columns:
            table.add_column(column)

        # Add rows to table
        for result in results:
            if isinstance(result, tuple):
                table.add_row(*(self.render(result, None, x) for x in result))
            else:
                table.add_row(*(self.render(result, column, result.get(column)) for column in columns))

        # Print table to console
        self.console.print(table)

    def workflow(self, command):
        """
        Processes .workflow command.

        Args:
            command: command line
        """

        command = shlex.split(command)
        if isinstance(self.app, Application):
            self.console.print(list(self.app.workflow(command[1], command[2:])))

    def isyaml(self, path):
        """
        Checks if file at path is a valid YAML file.

        Args:
            path: file to check

        Returns:
            True if file is valid YAML, False otherwise
        """

        if os.path.exists(path) and os.path.isfile(path):
            try:
                return Application.read(path)
            # pylint: disable=W0702
            except:
                pass

        return False

    def split(self, command, default=None):
        """
        Splits command by whitespace.

        Args:
            command: command line
            default: default command action

        Returns:
            command action
        """

        values = command.split(" ", 1)
        return values if len(values) > 1 else (command, default)

    def render(self, result, column, value):
        """
        Renders a search result column value.

        Args:
            result: result row
            column: column name
            value: column value
        """

        if isinstance(value, float):
            return f"{value:.4f}"

        # Explain highlighting
        if column == "text" and "tokens" in result:
            spans = []
            for token, score in result["tokens"]:
                color = None
                if score >= 0.02:
                    color = f"b {self.vhighlight}"

                spans.append((token, score, color))

            if result["score"] >= 0.05 and not [color for _, _, color in spans if color]:
                mscore = max(score for _, score, _ in spans)
                spans = [(token, score, f"b {self.vhighlight}" if score == mscore else color) for token, score, color in spans]

            output = ""
            for token, _, color in spans:
                if color:
                    output += f"[{color}]{token}[/{color}] "
                else:
                    output += f"{token} "

            return output

        return str(value)



"""
Task module
"""

import logging
import re
import types

import numpy as np
import torch

# Logging configuration
logger = logging.getLogger(__name__)


class Task:
    """
    Base class for all workflow tasks.
    """

    def __init__(
        self,
        action=None,
        select=None,
        unpack=True,
        column=None,
        merge="hstack",
        initialize=None,
        finalize=None,
        concurrency=None,
        onetomany=True,
        **kwargs,
    ):
        """
        Creates a new task. A task defines two methods, type of data it accepts and the action to execute
        for each data element. Action is a callable function or list of callable functions.

        Args:
            action: action(s) to execute on each data element
            select: filter(s) used to select data to process
            unpack: if data elements should be unpacked or unwrapped from (id, data, tag) tuples
            column: column index to select if element is a tuple, defaults to all
            merge: merge mode for joining multi-action outputs, defaults to hstack
            initialize: action to execute before processing
            finalize: action to execute after processing
            concurrency: sets concurrency method when execute instance available
                         valid values: "thread" for thread-based concurrency, "process" for process-based concurrency
            onetomany: if one-to-many data transformations should be enabled, defaults to True
            kwargs: additional keyword arguments
        """

        # Standardize into list of actions
        if not action:
            action = []
        elif not isinstance(action, list):
            action = [action]

        self.action = action
        self.select = select
        self.unpack = unpack
        self.column = column
        self.merge = merge
        self.initialize = initialize
        self.finalize = finalize
        self.concurrency = concurrency
        self.onetomany = onetomany

        # Check for custom registration. Adds additional instance members and validates required dependencies available.
        if hasattr(self, "register"):
            self.register(**kwargs)
        elif kwargs:
            # Raise error if additional keyword arguments passed in without register method
            kwargs = ", ".join(f"'{kw}'" for kw in kwargs)
            raise TypeError(f"__init__() got unexpected keyword arguments: {kwargs}")

    def __call__(self, elements, executor=None):
        """
        Executes action for a list of data elements.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        if isinstance(elements, list):
            return self.filteredrun(elements, executor)

        return self.run(elements, executor)

    def filteredrun(self, elements, executor):
        """
        Executes a filtered run, which will tag all inputs with a process id, filter elements down to elements the
        task can handle and execute on that subset. Items not selected for processing will be returned unmodified.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        # Build list of elements with unique process ids
        indexed = list(enumerate(elements))

        # Filter data down to data this task handles
        data = [(x, self.upack(element)) for x, element in indexed if self.accept(self.upack(element, True))]

        # Get list of filtered process ids
        ids = [x for x, _ in data]

        # Prepare elements and execute task action(s)
        results = self.execute([self.prepare(element) for _, element in data], executor)

        # Pack results back into elements
        if self.merge:
            elements = self.filteredpack(results, indexed, ids)
        else:
            elements = [self.filteredpack(r, indexed, ids) for r in results]

        return elements

    def filteredpack(self, results, indexed, ids):
        """
        Processes and packs results back into original input elements.

        Args:
            results: task results
            indexed: original elements indexed by process id
            ids: process ids accepted by this task

        Returns:
            packed elements
        """

        # Update with transformed elements. Handle one to many transformations.
        elements = []
        for x, element in indexed:
            if x in ids:
                # Get result for process id
                result = results[ids.index(x)]

                if isinstance(result, OneToMany):
                    # One to many transformations
                    elements.extend([self.pack(element, r) for r in result])
                else:
                    # One to one transformations
                    elements.append(self.pack(element, result))
            else:
                # Pass unprocessed elements through
                elements.append(element)

        return elements

    def run(self, elements, executor):
        """
        Executes a task run for elements. A standard run processes all elements.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        # Execute task actions
        results = self.execute(elements, executor)

        # Handle one to many transformations
        if isinstance(results, list):
            elements = []
            for result in results:
                if isinstance(result, OneToMany):
                    # One to many transformations
                    elements.extend(result)
                else:
                    # One to one transformations
                    elements.append(result)

            return elements

        return results

    def accept(self, element):
        """
        Determines if this task can handle the input data format.

        Args:
            element: input data element

        Returns:
            True if this task can process this data element, False otherwise
        """

        return (isinstance(element, str) and re.search(self.select, element.lower())) if element is not None and self.select else True

    def upack(self, element, force=False):
        """
        Unpacks data for processing.

        Args:
            element: input data element
            force: if True, data is unpacked even if task has unpack set to False

        Returns:
            data
        """

        # Extract data from (id, data, tag) formatted elements
        if (self.unpack or force) and isinstance(element, tuple) and len(element) > 1:
            return element[1]

        return element

    def pack(self, element, data):
        """
        Packs data after processing.

        Args:
            element: transformed data element
            data: item to pack element into

        Returns:
            packed data
        """

        # Pack data into (id, data, tag) formatted elements
        if self.unpack and isinstance(element, tuple) and len(element) > 1:
            # If new data is a (id, data, tag) tuple use that except for multi-action "hstack" merges which produce tuples
            if isinstance(data, tuple) and (len(self.action) <= 1 or self.merge != "hstack"):
                return data

            # Create a copy of tuple, update data element and return
            element = list(element)
            element[1] = data
            return tuple(element)

        return data

    def prepare(self, element):
        """
        Method that allows downstream tasks to prepare data element for processing.

        Args:
            element: input data element

        Returns:
            data element ready for processing
        """

        return element

    def execute(self, elements, executor):
        """
        Executes action(s) on elements.

        Args:
            elements: list of data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        if self.action:
            # Run actions
            outputs = []
            for x, action in enumerate(self.action):
                # Filter elements by column index if necessary - supports a single int or an action index to column index mapping
                index = self.column[x] if isinstance(self.column, dict) else self.column
                inputs = [self.extract(e, index) for e in elements] if index is not None else elements

                # Queue arguments for executor, process immediately if no executor available
                outputs.append((action, inputs) if executor else self.process(action, inputs))

            # Run with executor if available
            if executor:
                outputs = executor.run(self.concurrency, self.process, outputs)

            # Run post process operations
            return self.postprocess(outputs)

        return elements

    def extract(self, element, index):
        """
        Extracts a column from element by index if the element is a tuple.

        Args:
            element: input element
            index: column index

        Returns:
            extracted column
        """

        if isinstance(element, tuple):
            if not self.unpack and len(element) == 3 and isinstance(element[1], tuple):
                return (element[0], element[1][index], element[2])

            return element[index]

        return element

    def process(self, action, inputs):
        """
        Executes action using inputs as arguments.

        Args:
            action: callable object
            inputs: action inputs

        Returns:
            action outputs
        """

        # Log inputs
        logger.debug("Inputs: %s", inputs)

        # Execute action and get outputs
        outputs = action(inputs)

        # Consume generator output, if necessary
        if isinstance(outputs, types.GeneratorType):
            outputs = list(outputs)

        # Log outputs
        logger.debug("Outputs: %s", outputs)

        return outputs

    def postprocess(self, outputs):
        """
        Runs post process routines after a task action.

        Args:
            outputs: task outputs

        Returns:
            postprocessed outputs
        """

        # Unpack single action tasks
        if len(self.action) == 1:
            return self.single(outputs[0])

        # Return unmodified outputs when merge set to None
        if not self.merge:
            return outputs

        if self.merge == "vstack":
            return self.vstack(outputs)
        if self.merge == "concat":
            return self.concat(outputs)

        # Default mode is hstack
        return self.hstack(outputs)

    def single(self, outputs):
        """
        Post processes and returns single action outputs.

        Args:
            outputs: outputs from a single task

        Returns:
            post processed outputs
        """

        if self.onetomany and isinstance(outputs, list):
            # Wrap one to many transformations
            outputs = [OneToMany(output) if isinstance(output, list) else output for output in outputs]

        return outputs

    def vstack(self, outputs):
        """
        Merges outputs row-wise. Returns a list of lists which will be interpreted as a one to many transformation.

        Row-wise merge example (2 actions)

          Inputs: [a, b, c]

          Outputs => [[a1, b1, c1], [a2, b2, c2]]

          Row Merge => [[a1, a2], [b1, b2], [c1, c2]] = [a1, a2, b1, b2, c1, c2]

        Args:
            outputs: task outputs

        Returns:
            list of aggregated/zipped outputs as one to many transforms (row-wise)
        """

        # If all outputs are numpy arrays, use native method
        if all(isinstance(output, np.ndarray) for output in outputs):
            return np.concatenate(np.stack(outputs, axis=1))

        # If all outputs are torch tensors, use native method
        # pylint: disable=E1101
        if all(torch.is_tensor(output) for output in outputs):
            return torch.cat(tuple(torch.stack(outputs, axis=1)))

        # Flatten into lists of outputs per input row. Wrap as one to many transformation.
        merge = []
        for x in zip(*outputs):
            combine = []
            for y in x:
                if isinstance(y, list):
                    combine.extend(y)
                else:
                    combine.append(y)

            merge.append(OneToMany(combine))

        return merge

    def hstack(self, outputs):
        """
        Merges outputs column-wise. Returns a list of tuples which will be interpreted as a one to one transformation.

        Column-wise merge example (2 actions)

          Inputs: [a, b, c]

          Outputs => [[a1, b1, c1], [a2, b2, c2]]

          Column Merge => [(a1, a2), (b1, b2), (c1, c2)]

        Args:
            outputs: task outputs

        Returns:
            list of aggregated/zipped outputs as tuples (column-wise)
        """

        # If all outputs are numpy arrays, use native method
        if all(isinstance(output, np.ndarray) for output in outputs):
            return np.stack(outputs, axis=1)

        # If all outputs are torch tensors, use native method
        # pylint: disable=E1101
        if all(torch.is_tensor(output) for output in outputs):
            return torch.stack(outputs, axis=1)

        return list(zip(*outputs))

    def concat(self, outputs):
        """
        Merges outputs column-wise and concats values together into a string. Returns a list of strings.

        Concat merge example (2 actions)

          Inputs: [a, b, c]

          Outputs => [[a1, b1, c1], [a2, b2, c2]]

          Concat Merge => [(a1, a2), (b1, b2), (c1, c2)] => ["a1. a2", "b1. b2", "c1. c2"]

        Args:
            outputs: task outputs

        Returns:
            list of concat outputs
        """

        return [". ".join([str(y) for y in x if y]) for x in self.hstack(outputs)]


class OneToMany:
    """
    Encapsulates list output for a one to many transformation.
    """

    def __init__(self, values):
        """
        Creates a new OneToMany transformation.

        Args:
            values: list of outputs
        """

        self.values = values

    def __iter__(self):
        return self.values.__iter__()



"""
Hugging Face Transformers model wrapper module
"""

from ..models import Models
from .tensors import Tensors


class HFModel(Tensors):
    """
    Pipeline backed by a Hugging Face Transformers model.
    """

    def __init__(self, path=None, quantize=False, gpu=False, batch=64):
        """
        Creates a new HFModel.

        Args:
            path: optional path to model, accepts Hugging Face model hub id or local path,
                  uses default model for task if not provided
            quantize: if model should be quantized, defaults to False
            gpu: True/False if GPU should be enabled, also supports a GPU device id
            batch: batch size used to incrementally process content
        """

        # Default model path
        self.path = path

        # Quantization flag
        self.quantization = quantize

        # Get tensor device reference
        self.deviceid = Models.deviceid(gpu)
        self.device = Models.device(self.deviceid)

        # Process batch size
        self.batchsize = batch

    def prepare(self, model):
        """
        Prepares a model for processing. Applies dynamic quantization if necessary.

        Args:
            model: input model

        Returns:
            model
        """

        if self.deviceid == -1 and self.quantization:
            model = self.quantize(model)

        return model

    def tokenize(self, tokenizer, texts):
        """
        Tokenizes text using tokenizer. This method handles overflowing tokens and automatically splits
        them into separate elements. Indices of each element is returned to allow reconstructing the
        transformed elements after running through the model.

        Args:
            tokenizer: Tokenizer
            texts: list of text

        Returns:
            (tokenization result, indices)
        """

        # Pre-process and split on newlines
        batch, positions = [], []
        for x, text in enumerate(texts):
            elements = [t + " " for t in text.split("\n") if t]
            batch.extend(elements)
            positions.extend([x] * len(elements))

        # Run tokenizer
        tokens = tokenizer(batch, padding=True)

        inputids, attention, indices = [], [], []
        for x, ids in enumerate(tokens["input_ids"]):
            if len(ids) > tokenizer.model_max_length:
                # Remove padding characters, if any
                ids = [i for i in ids if i != tokenizer.pad_token_id]

                # Split into model_max_length chunks
                for chunk in self.batch(ids, tokenizer.model_max_length - 1):
                    # Append EOS token if necessary
                    if chunk[-1] != tokenizer.eos_token_id:
                        chunk.append(tokenizer.eos_token_id)

                    # Set attention mask
                    mask = [1] * len(chunk)

                    # Append padding if necessary
                    if len(chunk) < tokenizer.model_max_length:
                        pad = tokenizer.model_max_length - len(chunk)
                        chunk.extend([tokenizer.pad_token_id] * pad)
                        mask.extend([0] * pad)

                    inputids.append(chunk)
                    attention.append(mask)
                    indices.append(positions[x])
            else:
                inputids.append(ids)
                attention.append(tokens["attention_mask"][x])
                indices.append(positions[x])

        tokens = {"input_ids": inputids, "attention_mask": attention}

        # pylint: disable=E1102
        return ({name: self.tensor(tensor).to(self.device) for name, tensor in tokens.items()}, indices)



"""
DuckDB module
"""

import os
import re

from tempfile import TemporaryDirectory

# Conditional import
try:
    import duckdb

    DUCKDB = True
except ImportError:
    DUCKDB = False

from .embedded import Embedded
from .schema import Statement


class DuckDB(Embedded):
    """
    Database instance backed by DuckDB.
    """

    # Delete single document and object
    DELETE_DOCUMENT = "DELETE FROM documents WHERE id = ?"
    DELETE_OBJECT = "DELETE FROM objects WHERE id = ?"

    def __init__(self, config):
        super().__init__(config)

        if not DUCKDB:
            raise ImportError('DuckDB is not available - install "database" extra to enable')

    def execute(self, function, *args):
        # Call parent method with DuckDB compatible arguments
        return super().execute(function, *self.formatargs(args))

    def insertdocument(self, uid, data, tags, entry):
        # Delete existing document
        self.cursor.execute(DuckDB.DELETE_DOCUMENT, [uid])

        # Call parent method
        super().insertdocument(uid, data, tags, entry)

    def insertobject(self, uid, data, tags, entry):
        # Delete existing object
        self.cursor.execute(DuckDB.DELETE_OBJECT, [uid])

        # Call parent method
        super().insertobject(uid, data, tags, entry)

    def connect(self, path=":memory:"):
        # Create connection and start a transaction
        # pylint: disable=I1101
        connection = duckdb.connect(path)
        connection.begin()

        return connection

    def getcursor(self):
        return self.connection

    def jsonprefix(self):
        # Return json column prefix
        return "json_extract_string(data"

    def jsoncolumn(self, name):
        # Generate json column using json_extract function
        return f"json_extract_string(data, '$.{name}')"

    def rows(self):
        # Iteratively retrieve and yield rows
        batch = 256
        rows = self.cursor.fetchmany(batch)
        while rows:
            yield from rows
            rows = self.cursor.fetchmany(batch)

    def addfunctions(self):
        # DuckDB doesn't currently support scalar functions
        return

    def copy(self, path):
        # Delete existing file, if necessary
        if os.path.exists(path):
            os.remove(path)

        # Create database connection
        # pylint: disable=I1101
        connection = duckdb.connect(path)

        # List of tables
        tables = ["documents", "objects", "sections"]

        with TemporaryDirectory() as directory:
            # Export existing tables
            for table in tables:
                self.connection.execute(f"COPY {table} TO '{directory}/{table}.parquet' (FORMAT parquet)")

            # Create initial schema
            for schema in [Statement.CREATE_DOCUMENTS, Statement.CREATE_OBJECTS, Statement.CREATE_SECTIONS % "sections"]:
                connection.execute(schema)

            # Import tables into new schema
            for table in tables:
                connection.execute(f"COPY {table} FROM '{directory}/{table}.parquet' (FORMAT parquet)")

            # Create indexes and sync data to database file
            connection.execute(Statement.CREATE_SECTIONS_INDEX)
            connection.execute("CHECKPOINT")

        # Start transaction
        connection.begin()

        return connection

    def formatargs(self, args):
        """
        DuckDB doesn't support named parameters. This method replaces named parameters with question marks
        and makes parameters a list.

        Args:
            args: input arguments

        Returns:
            DuckDB compatible args
        """

        if args and len(args) > 1:
            # Unpack query args
            query, parameters = args

            # Iterate over parameters
            #   - Replace named parameters with ?'s
            #   - Build list of value with position indexes
            params = []
            for key, value in parameters.items():
                pattern = rf"\:{key}(?=\s|$)"
                match = re.search(pattern, query)
                if match:
                    query = re.sub(pattern, "?", query, count=1)
                    params.append((match.start(), value))

            # Repack query and parameter list
            args = (query, [value for _, value in sorted(params, key=lambda x: x[0])])

        return args



"""
Defines API paths for transcription endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/transcribe")
def transcribe(file: str):
    """
    Transcribes audio files to text.

    Args:
        file: file to transcribe

    Returns:
        transcribed text
    """

    return application.get().pipeline("transcription", (file,))


@router.post("/batchtranscribe")
def batchtranscribe(files: List[str] = Body(...)):
    """
    Transcribes audio files to text.

    Args:
        files: list of files to transcribe

    Returns:
        list of transcribed text
    """

    return application.get().pipeline("transcription", (files,))



"""
Main module.
"""

import sys

from .base import Console


def main(path=None):
    """
    Console execution loop.

    Args:
        path: model path
    """

    Console(path).cmdloop()


if __name__ == "__main__":
    main(sys.argv[1] if len(sys.argv) > 1 else None)



"""
Explain module
"""

import numpy as np


class Explain:
    """
    Explains the importance of each token in an input text element for a query. This method creates n permutations of the input text, where n
    is the number of tokens in the input text. This effectively masks each token to determine its importance to the query.
    """

    def __init__(self, embeddings):
        """
        Creates a new explain action.

        Args:
            embeddings: embeddings instance
        """

        self.embeddings = embeddings
        self.content = embeddings.config.get("content")

        # Alias embeddings attributes
        self.database = embeddings.database

    def __call__(self, queries, texts, limit):
        """
        Explains the importance of each input token in text for a list of queries.

        Args:
            query: input queries
            texts: optional list of (text|list of tokens), otherwise runs search queries
            limit: optional limit if texts is None

        Returns:
            list of dict per input text per query where a higher token scores represents higher importance relative to the query
        """

        # Construct texts elements per query
        texts = self.texts(queries, texts, limit)

        # Explain each query-texts combination
        return [self.explain(query, texts[x]) for x, query in enumerate(queries)]

    def texts(self, queries, texts, limit):
        """
        Constructs lists of dict for each input query.

        Args:
            queries: input queries
            texts: optional list of texts
            limit: optional limit if texts is None

        Returns:
            lists of dict for each input query
        """

        # Calculate similarity scores per query if texts present
        if texts:
            results = []
            for scores in self.embeddings.batchsimilarity(queries, texts):
                results.append([{"id": uid, "text": texts[uid], "score": score} for uid, score in scores])

            return results

        # Query for results if texts is None and content is enabled
        return self.embeddings.batchsearch(queries, limit) if self.content else [[]] * len(queries)

    def explain(self, query, texts):
        """
        Explains the importance of each input token in text for a list of queries.

        Args:
            query: input query
            texts: list of text

        Returns:
            list of {"id": value, "text": value, "score": value, "tokens": value} covering each input text element
        """

        # Explain results
        results = []

        # Parse out similar clauses, if necessary
        if self.database:
            # Parse query
            query = self.database.parse(query)

            # Extract query from similar clause
            query = " ".join([" ".join(clause) for clause in query["similar"]]) if "similar" in query else None

        # Return original texts if query, text or score not present
        if not query or not texts or "score" not in texts[0] or "text" not in texts[0]:
            return texts

        # Calculate result per input text element
        for result in texts:
            text = result["text"]
            tokens = text if isinstance(text, list) else text.split()

            # Create permutations of input text, masking each token to determine importance
            permutations = []
            for i in range(len(tokens)):
                data = tokens.copy()
                data.pop(i)
                permutations.append([" ".join(data)])

            # Calculate similarity for each input text permutation and get score delta as importance
            scores = [(i, result["score"] - np.abs(s)) for i, s in self.embeddings.similarity(query, permutations)]

            # Append tokens to result
            result["tokens"] = [(tokens[i], score) for i, score in sorted(scores, key=lambda x: x[0])]

            # Add data sorted in index order
            results.append(result)

        # Sort score descending and return
        return sorted(results, key=lambda x: x["score"], reverse=True)



"""
Defines API paths for caption endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/caption")
def caption(file: str):
    """
    Builds captions for images.

    Args:
        file: file to process

    Returns:
        list of captions
    """

    return application.get().pipeline("caption", (file,))


@router.post("/batchcaption")
def batchcaption(files: List[str] = Body(...)):
    """
    Builds captions for images.

    Args:
        files: list of files to process

    Returns:
        list of captions
    """

    return application.get().pipeline("caption", (files,))



"""
Route module
"""

from fastapi.routing import APIRoute, get_request_handler

from .responses import ResponseFactory


class EncodingAPIRoute(APIRoute):
    """
    Extended APIRoute that encodes responses based on HTTP Accept header.
    """

    def get_route_handler(self):
        """
        Resolves a response class based on the HTTP Accept header.

        Returns:
            route handler function
        """

        async def handler(request):
            route = get_request_handler(
                dependant=self.dependant,
                body_field=self.body_field,
                status_code=self.status_code,
                response_class=ResponseFactory.create(request),
                response_field=self.secure_cloned_response_field,
                response_model_include=self.response_model_include,
                response_model_exclude=self.response_model_exclude,
                response_model_by_alias=self.response_model_by_alias,
                response_model_exclude_unset=self.response_model_exclude_unset,
                response_model_exclude_defaults=self.response_model_exclude_defaults,
                response_model_exclude_none=self.response_model_exclude_none,
                dependency_overrides_provider=self.dependency_overrides_provider,
            )

            return await route(request)

        return handler



"""
Functions module
"""

from types import FunctionType, MethodType


class Functions:
    """
    Resolves function configuration to function references.
    """

    def __init__(self, embeddings):
        """
        Creates a new function resolver.

        Args:
            embeddings: embeddings instance
        """

        self.embeddings = embeddings

        # Handle to all reference objects
        self.references = None

    def __call__(self, config):
        """
        Resolves a list of functions to function references.

        Args:
            config: configuration

        Returns:
            list of function references
        """

        # Initialize stored references array
        self.references = []

        # Resolve callable functions
        functions = []
        for fn in config["functions"]:
            if isinstance(fn, dict):
                fn = fn.copy()
                fn["function"] = self.function(fn["function"])
            else:
                fn = self.function(fn)
            functions.append(fn)

        return functions

    def reset(self):
        """
        Clears all resolved references.
        """

        if self.references:
            for reference in self.references:
                reference.reset()

    def function(self, function):
        """
        Resolves function configuration. If function is a string, it's split on '.' and each part
        is separately resolved to an object, attribute or function. Each part is resolved upon the
        first invocation of the function. Otherwise, the input is returned.

        Args:
            function: function configuration

        Returns:
            function reference
        """

        if isinstance(function, str):
            parts = function.split(".")

            if hasattr(self.embeddings, parts[0]):
                m = Reference(self.embeddings, parts[0])
                self.references.append(m)
            else:
                module = ".".join(parts[:-1])
                m = __import__(module)

            for comp in parts[1:]:
                m = Reference(m, comp)
                self.references.append(m)

            return m

        return function


class Reference:
    """
    Stores a reference to an object attribute. This attribute is resolved by invoking the __call__ method.
    This allows for functions to be independent of the initialization order of an embeddings instance.
    """

    def __init__(self, obj, attribute):
        """
        Create a new reference.

        Args:
            obj: object handle
            attribute: attribute name
        """

        # Object handle and attribute
        self.obj = obj
        self.attribute = attribute

        # Keep a handle to the original inputs
        self.inputs = (obj, attribute)

        # True if the object and attribute have been resolved
        self.resolved = False

        # True if the attribute is a function
        self.function = None

    def __call__(self, *args):
        """
        Resolves an object attribute reference. If the attribute is a function, the function is executed.
        Otherwise, the object attribute value is returned.

        Args:
            args: list of function arguments to the object attribute, when attribute is a function

        Returns:
            object attribute function result or object attribute value
        """

        # Resolve nested function arguments, if necessary
        if not self.resolved:
            self.obj = self.obj() if isinstance(self.obj, Reference) else self.obj
            self.attribute = self.attribute() if isinstance(self.attribute, Reference) else self.attribute
            self.resolved = True

        # Lookup attribute
        attribute = getattr(self.obj, self.attribute)

        # Determine if attribute is a function
        if self.function is None:
            self.function = isinstance(attribute, (FunctionType, MethodType)) or (hasattr(attribute, "__call__") and args)

        # If attribute is a function, execute and return, otherwise return attribute
        return attribute(*args) if self.function else attribute

    def reset(self):
        """
        Clears resolved references.
        """

        self.obj, self.attribute = self.inputs
        self.resolved = False



"""
WorkflowTask module
"""

from .base import Task


class WorkflowTask(Task):
    """
    Task that executes a separate Workflow
    """

    def process(self, action, inputs):
        return list(super().process(action, inputs))



"""
ConsoleTask module
"""

import json

from .base import Task


class ConsoleTask(Task):
    """
    Task that prints task elements to the console.
    """

    def __call__(self, elements, executor=None):
        # Run task
        outputs = super().__call__(elements, executor)

        # Print inputs and outputs to console
        print("Inputs:", json.dumps(elements, indent=2))
        print("Outputs:", json.dumps(outputs, indent=2))

        # Return results
        return outputs



"""
AutoId module
"""

import inspect
import uuid


class AutoId:
    """
    Generates unique ids.
    """

    def __init__(self, method=None):
        """
        Creates a unique id generator.

        Args:
            method: generation method - supports int sequence (default) or UUID function
        """

        # Initialize variables
        self.method, self.function, self.value = None, None, None

        # Set id generation method
        if not method or isinstance(method, int):
            # Incrementing sequence (default)
            self.method = self.sequence
            self.value = method if method else 0
        else:
            # UUID generation function
            self.method = self.uuid
            self.function = getattr(uuid, method)

        # Check if signature takes a namespace argument (deterministic)
        args = inspect.getfullargspec(self.function).args if self.function else []
        self.deterministic = "namespace" in args

    def __call__(self, data=None):
        """
        Generates a unique id.

        Args:
            data: optional data to use for deterministic algorithms (i.e. uuid3, uuid5)

        Returns:
            unique id
        """

        return self.method(data)

    # pylint: disable=W0613
    def sequence(self, data):
        """
        Gets and increments sequence.

        Args:
            data: not used

        Returns:
            current sequence value
        """

        # Get and increment sequence
        value = self.value
        self.value += 1

        return value

    def uuid(self, data):
        """
        Generates a UUID and return as a string.

        Args:
            data: used with determistic algorithms (uuid3, uuid5)

        Returns:
            UUID string
        """

        uid = self.function(uuid.NAMESPACE_DNS, str(data)) if self.deterministic else self.function()
        return str(uid)

    def current(self):
        """
        Get the current sequence value. Only applicable for sequence ids, will be None for UUID methods.

        Returns:
            current sequence value
        """

        return self.value



"""
Defines API paths for extractor endpoints.
"""

from typing import List, Optional

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.post("/extract")
def extract(queue: List[dict] = Body(...), texts: Optional[List[str]] = Body(default=None)):
    """
    Extracts answers to input questions.

    Args:
        queue: list of {name: value, query: value, question: value, snippet: value}
        texts: optional list of text

    Returns:
        list of {name: value, answer: value}
    """

    return application.get().extract(queue, texts)



"""
Defines API paths for upload endpoints.
"""

import shutil
import tempfile

from typing import List

from fastapi import APIRouter, File, Form, UploadFile

from ..route import EncodingAPIRoute


router = APIRouter(route_class=EncodingAPIRoute)


@router.post("/upload")
def upload(files: List[UploadFile] = File(), suffix: str = Form(default=None)):
    """
    Uploads files for local server processing.

    Args:
        data: list of files to upload

    Returns:
        list of server paths
    """

    paths = []
    for f in files:
        with tempfile.NamedTemporaryFile(mode="wb", delete=False, suffix=suffix) as tmp:
            shutil.copyfileobj(f.file, tmp)
            paths.append(tmp.name)

    return paths



"""
Object storage module
"""

import os

# Conditional import
try:
    from libcloud.storage.providers import get_driver, DRIVERS
    from libcloud.storage.types import ContainerDoesNotExistError, ObjectDoesNotExistError

    LIBCLOUD = True
except ImportError:
    LIBCLOUD, DRIVERS = False, None


from .base import Cloud


class ObjectStorage(Cloud):
    """
    Object storage cloud provider backed by Apache libcloud.
    """

    @staticmethod
    def isprovider(provider):
        """
        Checks if this provider is an object storage provider.

        Args:
            provider: provider name

        Returns:
            True if this is an object storage provider
        """

        return LIBCLOUD and provider and provider.lower() in [x.lower() for x in DRIVERS]

    def __init__(self, config):
        super().__init__(config)

        if not LIBCLOUD:
            raise ImportError('Cloud object storage is not available - install "cloud" extra to enable')

        # Get driver for provider
        driver = get_driver(config["provider"])

        # Get client connection
        self.client = driver(
            config.get("key", os.environ.get("ACCESS_KEY")),
            config.get("secret", os.environ.get("ACCESS_SECRET")),
            **{field: config.get(field) for field in ["host", "port", "region", "token"] if config.get(field)},
        )

    def metadata(self, path=None):
        try:
            # If this is an archive path, check if file exists
            if self.isarchive(path):
                return self.client.get_object(self.config["container"], self.objectname(path))

            # Otherwise check if container exists
            return self.client.get_container(self.config["container"])
        except (ContainerDoesNotExistError, ObjectDoesNotExistError):
            return None

    def load(self, path=None):
        # Download archive file
        if self.isarchive(path):
            obj = self.client.get_object(self.config["container"], self.objectname(path))

            # Create local directory, if necessary
            directory = os.path.dirname(path)
            if directory:
                os.makedirs(directory, exist_ok=True)

            obj.download(path, overwrite_existing=True)

        # Download files in container. Optionally filter with a provided prefix.
        else:
            container = self.client.get_container(self.config["container"])
            for obj in container.list_objects(prefix=self.config.get("prefix")):
                # Derive local path and directory
                localpath = os.path.join(path, obj.name)
                directory = os.path.dirname(localpath)

                # Create local directory, if necessary
                os.makedirs(directory, exist_ok=True)

                # Download file locally
                obj.download(localpath, overwrite_existing=True)

        return path

    def save(self, path):
        # Get or create container
        try:
            container = self.client.get_container(self.config["container"])
        except ContainerDoesNotExistError:
            container = self.client.create_container(self.config["container"])

        # Upload files
        for f in self.listfiles(path):
            with open(f, "rb") as iterator:
                self.client.upload_object_via_stream(iterator=iterator, container=container, object_name=self.objectname(f))

    def objectname(self, name):
        """
        Derives an object name. This method checks if a prefix configuration parameter is present and combines
        it with the input name parameter.

        Args:
            name: input name

        Returns:
            object name
        """

        # Get base name
        name = os.path.basename(name)

        # Get optional prefix/folder
        prefix = self.config.get("prefix")

        # Prepend prefix, if applicable
        return f"{prefix}/{name}" if prefix else name



"""
RDBMS module
"""

import datetime
import json

from .base import Database
from .schema import Statement


# pylint: disable=R0904
class RDBMS(Database):
    """
    Base relational database class. A relational database uses SQL to insert, update, delete and select from a
    database instance.
    """

    def __init__(self, config):
        """
        Creates a new Database.

        Args:
            config: database configuration parameters
        """

        super().__init__(config)

        # Database connection
        self.connection = None
        self.cursor = None

    def load(self, path):
        # Load an existing database. Thread locking must be handled externally.
        self.session(path)

    def insert(self, documents, index=0):
        # Initialize connection if not open
        self.initialize()

        # Get entry date
        entry = datetime.datetime.now(datetime.timezone.utc)

        # Insert documents
        for uid, document, tags in documents:
            if isinstance(document, dict):
                # Insert document and use return value for sections table
                document = self.loaddocument(uid, document, tags, entry)

            if document is not None:
                if isinstance(document, list):
                    # Join tokens to text
                    document = " ".join(document)
                elif not isinstance(document, str):
                    # If object support is enabled, save object
                    self.loadobject(uid, document, tags, entry)

                    # Clear section text for objects, even when objects aren't inserted
                    document = None

                # Save text section
                self.loadsection(index, uid, document, tags, entry)
                index += 1

        # Post processing logic
        self.finalize()

    def delete(self, ids):
        if self.connection:
            # Batch ids
            self.batch(ids=ids)

            # Delete all documents, objects and sections by id
            self.cursor.execute(Statement.DELETE_DOCUMENTS)
            self.cursor.execute(Statement.DELETE_OBJECTS)
            self.cursor.execute(Statement.DELETE_SECTIONS)

    def reindex(self, config):
        if self.connection:
            # Set new configuration
            self.configure(config)

            # Resolve text column
            select = self.resolve(self.text)

            # Initialize reindex operation
            name = self.reindexstart()

            # Copy data over
            self.cursor.execute(Statement.COPY_SECTIONS % (name, select))

            # Stream new results
            self.cursor.execute(Statement.STREAM_SECTIONS % name)
            for uid, text, data, obj, tags in self.rows():
                if not text and self.encoder and obj:
                    yield (uid, self.encoder.decode(obj), tags)
                else:
                    # Read JSON data, if provided
                    data = json.loads(data) if data and isinstance(data, str) else data

                    # Stream data if available, otherwise use section text
                    yield (uid, data if data else text, tags)

            # Swap as new table
            self.cursor.execute(Statement.DROP_SECTIONS)
            self.cursor.execute(Statement.RENAME_SECTIONS % name)

            # Finish reindex operation
            self.reindexend(name)

    def save(self, path):
        if self.connection:
            self.connection.commit()

    def close(self):
        # Close connection
        if self.connection:
            self.connection.close()

    def ids(self, ids):
        # Batch ids and run query
        self.batch(ids=ids)
        self.cursor.execute(Statement.SELECT_IDS)

        # Format and return results
        return self.cursor.fetchall()

    def count(self):
        self.cursor.execute(Statement.COUNT_IDS)
        return self.cursor.fetchone()[0]

    def resolve(self, name, alias=None):
        # Standard column names
        sections = ["indexid", "id", "tags", "entry"]
        noprefix = ["data", "object", "score", "text"]

        # Alias expression
        if alias:
            # Skip if name matches alias or alias is a standard column name
            if name == alias or alias in sections:
                return name

            # Build alias clause
            return f'{name} as "{alias}"'

        # Resolve expression
        if self.expressions and name in self.expressions:
            return self.expressions[name]

        # Name is already resolved, skip
        if name.startswith(self.jsonprefix()) or any(f"s.{s}" == name for s in sections):
            return name

        # Standard columns - need prefixes
        if name.lower() in sections:
            return f"s.{name}"

        # Standard columns - no prefixes
        if name.lower() in noprefix:
            return name

        # Other columns come from documents.data JSON
        return self.jsoncolumn(name)

    def embed(self, similarity, batch):
        # Load similarity results id batch
        self.batch(indexids=[i for i, _ in similarity[batch]], batch=batch)

        # Average and load all similarity scores with first batch
        if not batch:
            self.scores(similarity)

        # Return ids clause placeholder
        return Statement.IDS_CLAUSE % batch

    # pylint: disable=R0912
    def query(self, query, limit, parameters, indexids):
        # Extract query components
        select = query.get("select", self.defaults())
        where = query.get("where")
        groupby, having = query.get("groupby"), query.get("having")
        orderby, qlimit, offset = query.get("orderby"), query.get("limit"), query.get("offset")
        similarity = query.get("similar")

        # Select "indexid, score" when indexids is True
        if indexids:
            select = f"{self.resolve('indexid')}, {self.resolve('score')}"

        # Build query text
        query = Statement.TABLE_CLAUSE % select
        if where is not None:
            query += f" WHERE {where}"
        if groupby is not None:
            query += f" GROUP BY {groupby}"
        if having is not None:
            query += f" HAVING {having}"
        if orderby is not None:
            query += f" ORDER BY {orderby}"

        # Default ORDER BY if not provided and similarity scores are available
        if similarity and orderby is None:
            query += " ORDER BY score DESC"

        # Apply query limit
        if qlimit is not None or limit:
            query += f" LIMIT {qlimit if qlimit else limit}"

            # Apply offset
            if offset is not None:
                query += f" OFFSET {offset}"

        # Clear scores when no similar clauses present
        if not similarity:
            self.scores(None)

        # Runs a user query through execute method, which has common user query handling logic
        args = (query, parameters) if parameters else (query,)
        self.execute(self.cursor.execute, *args)

        # Retrieve column list from query
        columns = [c[0] for c in self.cursor.description]

        # Map results and return
        results = []
        for row in self.rows():
            result = {}

            # Copy columns to result. In cases with duplicate column names, find one with a value
            for x, column in enumerate(columns):
                if column not in result or result[column] is None:
                    # Decode object
                    if self.encoder and column == self.object:
                        result[column] = self.encoder.decode(row[x])
                    else:
                        result[column] = row[x]

            results.append(result)

        # Transform results, if necessary
        return [(x["indexid"], x["score"]) for x in results] if indexids else results

    def initialize(self):
        """
        Creates connection and initial database schema if no connection exists.
        """

        if not self.connection:
            # Create database session. Thread locking must be handled externally.
            self.session()

            # Create initial table schema
            self.createtables()

    def session(self, path=None, connection=None):
        """
        Starts a new database session.

        Args:
            path: path to database file
            connection: existing connection to use
        """

        # Create database connection and cursor
        self.connection = connection if connection else self.connect(path) if path else self.connect()
        self.cursor = self.getcursor()

        # Register custom functions - session scope
        self.addfunctions()

        # Create temporary tables - session scope
        self.createbatch()
        self.createscores()

    def createtables(self):
        """
        Creates the initial table schema.
        """

        self.cursor.execute(Statement.CREATE_DOCUMENTS)
        self.cursor.execute(Statement.CREATE_OBJECTS)
        self.cursor.execute(Statement.CREATE_SECTIONS % "sections")
        self.cursor.execute(Statement.CREATE_SECTIONS_INDEX)

    def finalize(self):
        """
        Post processing logic run after inserting a batch of documents. Default method is no-op.
        """

    def loaddocument(self, uid, document, tags, entry):
        """
        Applies pre-processing logic and inserts a document.

        Args:
            uid: unique id
            document: input document dictionary
            tags: document tags
            entry: generated entry date

        Returns:
            section value
        """

        # Make a copy of document before changing
        document = document.copy()

        # Get and remove object field from document
        obj = document.pop(self.object) if self.object in document else None

        # Insert document as JSON
        if document:
            self.insertdocument(uid, json.dumps(document, allow_nan=False), tags, entry)

        # If text and object are both available, load object as it won't otherwise be used
        if self.text in document and obj:
            self.loadobject(uid, obj, tags, entry)

        # Return value to use for section - use text if available otherwise use object
        return document[self.text] if self.text in document else obj

    def insertdocument(self, uid, data, tags, entry):
        """
        Inserts a document.

        Args:
            uid: unique id
            data: document data
            tags: document tags
            entry: generated entry date
        """

        self.cursor.execute(Statement.INSERT_DOCUMENT, [uid, data, tags, entry])

    def loadobject(self, uid, obj, tags, entry):
        """
        Applies pre-preprocessing logic and inserts an object.

        Args:
            uid: unique id
            obj: input object
            tags: object tags
            entry: generated entry date
        """

        # If object support is enabled, save object
        if self.encoder:
            self.insertobject(uid, self.encoder.encode(obj), tags, entry)

    def insertobject(self, uid, data, tags, entry):
        """
        Inserts an object.

        Args:
            uid: unique id
            data: encoded data
            tags: object tags
            entry: generated entry date
        """

        self.cursor.execute(Statement.INSERT_OBJECT, [uid, data, tags, entry])

    def loadsection(self, index, uid, text, tags, entry):
        """
        Applies pre-processing logic and inserts a section.

        Args:
            index: index id
            uid: unique id
            text: section text
            tags: section tags
            entry: generated entry date
        """

        self.insertsection(index, uid, text, tags, entry)

    def insertsection(self, index, uid, text, tags, entry):
        """
        Inserts a section.

        Args:
            index: index id
            uid: unique id
            text: section text
            tags: section tags
            entry: generated entry date
        """

        # Save text section
        self.cursor.execute(Statement.INSERT_SECTION, [index, uid, text, tags, entry])

    def reindexstart(self):
        """
        Starts a reindex operation.

        Returns:
            temporary working table name
        """

        # Working table name
        name = "rebuild"

        # Create new table to hold reordered sections
        self.cursor.execute(Statement.CREATE_SECTIONS % name)

        return name

    # pylint: disable=W0613
    def reindexend(self, name):
        """
        Ends a reindex operation.

        Args:
            name: working table name
        """

        self.cursor.execute(Statement.CREATE_SECTIONS_INDEX)

    def batch(self, indexids=None, ids=None, batch=None):
        """
        Loads ids to a temporary batch table for efficient query processing.

        Args:
            indexids: list of indexids
            ids: list of ids
            batch: batch index, used when statement has multiple subselects
        """

        # Delete batch when batch id is empty or for batch 0
        if not batch:
            self.cursor.execute(Statement.DELETE_BATCH)

        # Add batch
        self.insertbatch(indexids, ids, batch)

    def createbatch(self):
        """
        Creates temporary batch table.
        """

        # Create or Replace temporary batch table
        self.cursor.execute(Statement.CREATE_BATCH)

    def insertbatch(self, indexids, ids, batch):
        """
        Inserts batch of ids.
        """

        if indexids:
            self.cursor.executemany(Statement.INSERT_BATCH_INDEXID, [(i, batch) for i in indexids])
        if ids:
            self.cursor.executemany(Statement.INSERT_BATCH_ID, [(str(uid), batch) for uid in ids])

    def scores(self, similarity):
        """
        Loads a batch of similarity scores to a temporary table for efficient query processing.

        Args:
            similarity: similarity results as [(indexid, score)]
        """

        # Delete scores
        self.cursor.execute(Statement.DELETE_SCORES)

        if similarity:
            # Average scores per id, needed for multiple similar() clauses
            scores = {}
            for s in similarity:
                for i, score in s:
                    if i not in scores:
                        scores[i] = []
                    scores[i].append(score)

            # Add scores
            self.insertscores(scores)

    def createscores(self):
        """
        Creates temporary scores table.
        """

        # Create or Replace temporary scores table
        self.cursor.execute(Statement.CREATE_SCORES)

    def insertscores(self, scores):
        """
        Inserts a batch of scores.

        Args:
            scores: scores to add
        """

        # Average scores by id
        if scores:
            self.cursor.executemany(Statement.INSERT_SCORE, [(i, sum(s) / len(s)) for i, s in scores.items()])

    def defaults(self):
        """
        Returns a list of default columns when there is no select clause.

        Returns:
            list of default columns
        """

        return "s.id, text, score"

    def connect(self, path=None):
        """
        Creates a new database connection.

        Args:
            path: path to database file

        Returns:
            connection
        """

        raise NotImplementedError

    def getcursor(self):
        """
        Opens a cursor for current connection.

        Returns:
            cursor
        """

        raise NotImplementedError

    def jsonprefix(self):
        """
        Returns json column prefix to test for.

        Returns:
            dynamic column prefix
        """

        raise NotImplementedError

    def jsoncolumn(self, name):
        """
        Builds a json extract column expression for name.

        Args:
            name: column name

        Returns:
            dynamic column expression
        """

        raise NotImplementedError

    def rows(self):
        """
        Returns current cursor row iterator for last executed query.

        Args:
            cursor: cursor

        Returns:
            iterable collection of rows
        """

        raise NotImplementedError

    def addfunctions(self):
        """
        Adds custom functions in current connection.
        """

        raise NotImplementedError



"""
RAG module
"""

from ...models import Models

from ..base import Pipeline
from ..data import Tokenizer
from ..text import Questions
from ..text import Similarity

from .factory import GenerationFactory
from .llm import LLM


class RAG(Pipeline):
    """
    Extracts knowledge from content by joining a prompt, context data store and generative model together. The data store can be
    an embeddings database or a similarity instance with associated input text. The generative model can be a prompt-driven large
    language model (LLM), an extractive question-answering model or a custom pipeline. This is known as retrieval augmented generation (RAG).
    """

    # pylint: disable=R0913
    def __init__(
        self,
        similarity,
        path,
        quantize=False,
        gpu=True,
        model=None,
        tokenizer=None,
        minscore=None,
        mintokens=None,
        context=None,
        task=None,
        output="default",
        template=None,
        separator=" ",
        system=None,
        **kwargs,
    ):
        """
        Builds a new RAG pipeline.

        Args:
            similarity: similarity instance (embeddings or similarity pipeline)
            path: path to model, supports a LLM, Questions or custom pipeline
            quantize: True if model should be quantized before inference, False otherwise.
            gpu: if gpu inference should be used (only works if GPUs are available)
            model: optional existing pipeline model to wrap
            tokenizer: Tokenizer class
            minscore: minimum score to include context match, defaults to None
            mintokens: minimum number of tokens to include context match, defaults to None
            context: topn context matches to include, defaults to 3
            task: model task (language-generation, sequence-sequence or question-answering), defaults to auto-detect
            output: output format, 'default' returns (name, answer), 'flatten' returns answers and 'reference' returns (name, answer, reference)
            template: prompt template, it must have a parameter for {question} and {context}, defaults to "{question} {context}"
            separator: context separator
            system: system prompt, defaults to None
            kwargs: additional keyword arguments to pass to pipeline model
        """

        # Similarity instance
        self.similarity = similarity

        # Model can be a LLM, Questions or custom pipeline
        self.model = self.load(path, quantize, gpu, model, task, **kwargs)

        # Tokenizer class use default method if not set
        self.tokenizer = tokenizer if tokenizer else Tokenizer() if hasattr(self.similarity, "scoring") and self.similarity.isweighted() else None

        # Minimum score to include context match
        self.minscore = minscore if minscore is not None else 0.0

        # Minimum number of tokens to include context match
        self.mintokens = mintokens if mintokens is not None else 0.0

        # Top n context matches to include for context
        self.context = context if context else 3

        # Output format
        self.output = output

        # Prompt template
        self.template = template if template else "{question} {context}"

        # Context separator
        self.separator = separator

        # System prompt template
        self.system = system

    def __call__(self, queue, texts=None, **kwargs):
        """
        Finds answers to input questions. This method runs queries to find the top n best matches and uses that as the context.
        A model is then run against the context for each input question, with the answer returned.

        Args:
            queue: input question queue (name, query, question, snippet), can be list of tuples/dicts/strings or a single input element
            texts: optional list of text for context, otherwise runs embeddings search
            kwargs: additional keyword arguments to pass to pipeline model

        Returns:
            list of answers matching input format (tuple or dict) containing fields as specified by output format
        """

        # Save original queue format
        inputs = queue

        # Convert queue to list, if necessary
        queue = queue if isinstance(queue, list) else [queue]

        # Convert dictionary inputs to tuples
        if queue and isinstance(queue[0], dict):
            # Convert dict to tuple
            queue = [tuple(row.get(x) for x in ["name", "query", "question", "snippet"]) for row in queue]

        if queue and isinstance(queue[0], str):
            # Convert string questions to tuple
            queue = [(None, row, row, None) for row in queue]

        # Rank texts by similarity for each query
        results = self.query([query for _, query, _, _ in queue], texts)

        # Build question-context pairs
        names, queries, questions, contexts, topns, snippets = [], [], [], [], [], []
        for x, (name, query, question, snippet) in enumerate(queue):
            # Get top n best matching segments
            topn = sorted(results[x], key=lambda y: y[2], reverse=True)[: self.context]

            # Generate context using ordering from texts, if available, otherwise order by score
            context = self.separator.join(text for _, text, _ in (sorted(topn, key=lambda y: y[0]) if texts else topn))

            names.append(name)
            queries.append(query)
            questions.append(question)
            contexts.append(context)
            topns.append(topn)
            snippets.append(snippet)

        # Run pipeline and return answers
        answers = self.answers(questions, contexts, **kwargs)

        # Apply output formatting to answers and return
        return self.apply(inputs, names, queries, answers, topns, snippets) if isinstance(answers, list) else answers

    def load(self, path, quantize, gpu, model, task, **kwargs):
        """
        Loads a LLM, Questions or custom pipeline.

        Args:
            path: path to model, supports a LLM, Questions or custom pipeline
            quantize: True if model should be quantized before inference, False otherwise.
            gpu: if gpu inference should be used (only works if GPUs are available)
            model: optional existing pipeline model to wrap
            task: model task (language-generation, sequence-sequence or question-answering), defaults to auto-detect
            kwargs: additional keyword arguments to pass to pipeline model

        Returns:
            LLM, Questions or custom pipeline
        """

        # Only try to load if path is a string
        if not isinstance(path, str):
            return path

        # Attempt to resolve task if not provided
        task = GenerationFactory.method(path, task)
        task = Models.task(path, **kwargs) if task == "transformers" else task

        # Load Questions pipeline
        if task == "question-answering":
            return Questions(path, quantize, gpu, model, **kwargs)

        # Load LLM pipeline
        return LLM(path=path, quantize=quantize, gpu=gpu, model=model, task=task, **kwargs)

    def query(self, queries, texts):
        """
        Rank texts by similarity for each query. If texts is empty, an embeddings search will be executed.
        Returns results sorted by best match.

        Args:
            queries: list of queries
            texts: optional list of text

        Returns:
            list of (id, data, score) per query
        """

        if not queries:
            return []

        # Score text against queries
        scores, segments, tokenlist = self.score(queries, texts)

        # Build question-context pairs
        results = []
        for i, query in enumerate(queries):
            # Get list of required and prohibited tokens
            must = [token.strip("+") for token in query.split() if token.startswith("+") and len(token) > 1]
            mnot = [token.strip("-") for token in query.split() if token.startswith("-") and len(token) > 1]

            # Segment text is static when texts is passed in but different per query when an embeddings search is run
            segment = segments if texts else segments[i]
            tokens = tokenlist if texts else tokenlist[i]

            # List of matches
            matches = []
            for y, (x, score) in enumerate(scores[i]):
                # Segments and tokens are statically ordered when texts is passed in, need to resolve values with score id
                # Scores, segments and tokens all share the same list ordering when an embeddings search is run
                x = x if texts else y

                # Get segment text
                text = segment[x][1]

                # Add result if:
                #   - all required tokens are present or there are not required tokens AND
                #   - all prohibited tokens are not present or there are not prohibited tokens
                #   - score is above minimum score required
                #   - number of tokens is above minimum number of tokens required
                if (not must or all(token.lower() in text.lower() for token in must)) and (
                    not mnot or all(token.lower() not in text.lower() for token in mnot)
                ):
                    if score >= self.minscore and len(tokens[x]) >= self.mintokens:
                        matches.append(segment[x] + (score,))

            # Add query matches sorted by highest score
            results.append(matches)

        return results

    def score(self, queries, texts):
        """
        Runs queries against texts (or an embeddings search if texts is empty) and builds list of
        similarity scores for each query-text combination.

        Args:
            queries: list of queries
            texts: optional list of text

        Returns:
            scores, segments, tokenlist
        """

        # Tokenize text
        segments, tokenlist = [], []
        if texts:
            for text in texts:
                # Run tokenizer method, if available, otherwise returns original text
                tokens = self.tokenize(text)
                if tokens:
                    segments.append(text)
                    tokenlist.append(tokens)

            # Add index id to segments to preserve ordering after filters
            segments = list(enumerate(segments))

        # Get list of (id, score) - sorted by highest score per query
        if isinstance(self.similarity, Similarity):
            # Score using similarity pipeline
            scores = self.similarity(queries, [t for _, t in segments])
        elif texts:
            # Score using embeddings.batchsimilarity
            scores = self.similarity.batchsimilarity([self.tokenize(x) for x in queries], tokenlist)
        else:
            # Score using embeddings.batchsearch
            scores, segments, tokenlist = self.batchsearch(queries)

        return scores, segments, tokenlist

    def batchsearch(self, queries):
        """
        Runs a batch embeddings search for a set of queries.

        Args:
            queries: list of queries to run

        Returns:
            scores, segments, tokenlist
        """

        scores, segments, tokenlist = [], [], []
        for results in self.similarity.batchsearch([self.tokenize(x) for x in queries], self.context):
            # Assume embeddings content is enabled and results are dictionaries
            scores.append([(result["id"], result["score"]) for result in results])
            segments.append([(result["id"], result["text"]) for result in results])
            tokenlist.append([self.tokenize(result["text"]) for result in results])

        return scores, segments, tokenlist

    def tokenize(self, text):
        """
        Tokenizes text. Returns original text if tokenizer is not available.

        Args:
            text: input text

        Returns:
            tokens if tokenizer available otherwise original text
        """

        return self.tokenizer(text) if self.tokenizer else text

    def answers(self, questions, contexts, **kwargs):
        """
        Executes pipeline and formats extracted answers.

        Args:
            questions: questions
            contexts: question context
            kwargs: additional keyword arguments to pass to model

        Returns:
            answers
        """

        # Run model inference with questions pipeline
        if isinstance(self.model, Questions):
            return self.model(questions, contexts)

        # Run generator pipeline
        return self.model(self.prompts(questions, contexts), **kwargs)

    def prompts(self, questions, contexts):
        """
        Builds a list of prompts using the passed in questions and contexts.

        Args:
            questions: questions
            contexts: question context

        Returns:
            prompts
        """

        # Format prompts for generator pipeline
        prompts = []
        for x, context in enumerate(contexts):
            # Create input prompt
            prompt = self.template.format(question=questions[x], context=context)

            # Add system prompt, if necessary
            if self.system:
                prompt = [
                    {"role": "system", "content": self.system.format(question=questions[x], context=context)},
                    {"role": "user", "content": prompt},
                ]

            prompts.append(prompt)

        return prompts

    def apply(self, inputs, names, queries, answers, topns, snippets):
        """
        Applies the following formatting rules to answers.
            - each answer row matches input format (tuple or dict)
            - if output format is 'flatten' then this method flattens to a list of answers
            - if output format is 'reference' then a list of (name, answer, reference) is returned
            - otherwise, if output format is 'default' or anything else list of (name, answer) is returned

        Args:
            inputs: original inputs
            names: question identifiers/names
            queries: list of input queries
            answers: list of generated answers
            topns: top n records used for context
            snippets: flags to enable answer snippets per answer

        Returns:
            list of answers matching input format (tuple or dict) containing fields as specified by output format
        """

        # Resolve answers as snippets
        answers = self.snippets(names, answers, topns, snippets)

        # Flatten to list of answers and return
        if self.output == "flatten":
            answers = [answer for _, answer in answers]
        else:
            # Resolve id reference for each answer
            if self.output == "reference":
                answers = self.reference(queries, answers, topns)

            # Ensure output format matches input format
            first = inputs[0] if inputs and isinstance(inputs, list) else inputs
            if isinstance(first, (dict, str)):
                # Add name if input queue had name field
                fields = ["name", "answer", "reference"] if isinstance(first, dict) and "name" in first else [None, "answer", "reference"]
                answers = [{fields[x]: column for x, column in enumerate(row) if fields[x]} for row in answers]

        # Unpack single answer, if necessary
        return answers[0] if answers and isinstance(inputs, (tuple, dict, str)) else answers

    def snippets(self, names, answers, topns, snippets):
        """
        Extracts text surrounding the answer within context.

        Args:
            names: question identifiers/names
            answers: list of generated answers
            topns: top n records used for context
            snippets: flags to enable answer snippets per answer

        Returns:
            answers resolved as snippets per question, if necessary
        """

        # Extract and format answer
        results = []

        for x, answer in enumerate(answers):
            # Resolve snippet if necessary
            if answer and snippets[x]:
                # Searches for first text element to contain answer
                for _, text, _ in topns[x]:
                    if answer in text:
                        answer = text
                        break

            results.append((names[x], answer))

        return results

    def reference(self, queries, answers, topns):
        """
        Reference each answer with the best matching context element id.

        Args:
            queries: list of input queries
            answers: list of answers
            topn: top n context elements as (id, data, tag)

        Returns:
            list of (name, answer, reference)
        """

        # Convert queries to terms
        terms = self.terms(queries)

        outputs = []
        for x, (name, answer) in enumerate(answers):
            # Get matching topn
            topn, reference = topns[x], None

            if topn:
                # Build query from keyword terms and the answer text
                query = f"{terms[x]} {answers[x][1]}"

                # Compare answer to topns to find best match
                scores, _, _ = self.score([query], [text for _, text, _ in topn])

                # Get top score index
                index = scores[0][0][0]

                # Use matching topn id as reference
                reference = topn[index][0]

            # Append (name, answer, reference) tuple
            outputs.append((name, answer, reference))

        return outputs

    def terms(self, queries):
        """
        Extracts keyword terms from a list of queries using underlying similarity model.

        Args:
            queries: list of queries

        Returns:
            list of queries reduced down to keyword term strings
        """

        # Extract keyword terms from queries if underlying similarity model supports it
        return self.similarity.batchterms(queries) if hasattr(self.similarity, "batchterms") else queries



"""
Tar module
"""

import os
import tarfile

from .compress import Compress


class Tar(Compress):
    """
    Tar compression
    """

    def pack(self, path, output):
        # Infer compression type
        compression = self.compression(output)

        with tarfile.open(output, f"w:{compression}" if compression else "w") as tar:
            tar.add(path, arcname=".")

    def unpack(self, path, output):
        # Infer compression type
        compression = self.compression(path)

        with tarfile.open(path, f"r:{compression}" if compression else "r") as tar:
            # Validate paths
            for member in tar.getmembers():
                fullpath = os.path.join(path, member.name)
                if not self.validate(path, fullpath):
                    raise IOError(f"Invalid tar entry: {member.name}")

            tar.extractall(output)

    def compression(self, path):
        """
        Gets compression type for path.

        Args:
            path: path to file

        Returns:
            compression type
        """

        # Infer compression type from last path component. Limit to supported types.
        compression = path.lower().split(".")[-1]
        return compression if compression in ("bz2", "gz", "xz") else None



"""
API factory module
"""

from ..util import Resolver


class APIFactory:
    """
    API factory. Creates new API instances.
    """

    @staticmethod
    def get(api):
        """
        Gets a new instance of api class.

        Args:
            api: API instance class

        Returns:
            API
        """

        return Resolver()(api)

    @staticmethod
    def create(config, api):
        """
        Creates a new API instance.

        Args:
            config: API configuration
            api: API instance class

        Returns:
            API instance
        """

        return APIFactory.get(api)(config)



"""
AudioStream module
"""

from queue import Queue
from threading import Thread

# Conditional import
try:
    import sounddevice as sd

    from .signal import Signal, SCIPY

    AUDIOSTREAM = SCIPY
except (ImportError, OSError):
    AUDIOSTREAM = False

from ..base import Pipeline


class AudioStream(Pipeline):
    """
    Threaded pipeline that streams audio segments to an output audio device. This pipeline is designed
    to run on local machines given that it requires access to write to an output device.
    """

    # End of stream message
    COMPLETE = (1, None)

    def __init__(self, rate=None):
        """
        Creates an AudioStream pipeline.

        Args:
            rate: optional target sample rate, otherwise uses input target rate with each audio segment
        """

        if not AUDIOSTREAM:
            raise ImportError(
                (
                    'AudioStream pipeline is not available - install "pipeline" extra to enable. '
                    "Also check that the portaudio system library is available."
                )
            )

        # Target sample rate
        self.rate = rate

        self.queue = Queue()
        self.thread = Thread(target=self.play)
        self.thread.start()

    def __call__(self, segment):
        """
        Queues audio segments for the audio player.

        Args:
            segment: (audio, sample rate)|list

        Returns:
            segment
        """

        # Convert single element to list
        segments = [segment] if isinstance(segment, tuple) else segment

        for x in segments:
            self.queue.put(x)

        # Return single element if single element passed in
        return segments[0] if isinstance(segment, tuple) else segments

    def wait(self):
        """
        Waits for all input audio segments to be played.
        """

        self.thread.join()

    def play(self):
        """
        Reads audio segments from queue. This method runs in a separate non-blocking thread.
        """

        audio, rate = self.queue.get()
        while not isinstance(audio, int) or (audio, rate) != AudioStream.COMPLETE:
            # Resample to target sample rate, if necessary
            audio, rate = (Signal.resample(audio, rate, self.rate), self.rate) if self.rate else (audio, rate)

            # Play audio segment
            sd.play(audio, rate, blocking=True)

            # Get next segment
            audio, rate = self.queue.get()



"""
Defines API paths for tabular endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/tabular")
def tabular(file: str):
    """
    Splits tabular data into rows and columns.

    Args:
        file: file to process

    Returns:
        list of (id, text, tag) elements
    """

    return application.get().pipeline("tabular", (file,))


@router.post("/batchtabular")
def batchtabular(files: List[str] = Body(...)):
    """
    Splits tabular data into rows and columns.

    Args:
        files: list of files to process

    Returns:
        list of (id, text, tag) elements
    """

    return application.get().pipeline("tabular", (files,))
